import math

import numpy as np

import deampy.random_variates as RVGs


class OneDimDataFrame:
    """
    example:
        age,    mortality rate
        0,      0.1,
        5,      0.2,
        10,     0.3
    """
    def __init__(self, y_objects, x_min=0, x_max=1, x_delta=1):
        """
        :param y_objects: (list) of y objects (in example above: [0.1, 0.2, 0.3])
                but it could also be a list of any objects.
        :param x_min: minimum value of x (in example above: 0)
        :param x_max: maximum value of x (in example above: 10)
        :param x_delta: interval between break points of x (in example above: 5)
                    if set to 'int', x is treated as categorical variable
        """

        self.objs = y_objects
        self.xDelta = x_delta
        self.xMin = x_min
        self.xMax = x_max

    def get_index(self, x_value):
        """ :returns the index of the smallest x break point with x_value greater than or equal.
            In the example above, it returns
                0 for 0 <= x_value < 5,
                1 for 5 <= x_value < 10, and
                2 for x_value >= 10
        """

        if self.xDelta == 'int':
            if type(x_value) is not int:
                raise ValueError('x_value should be an integer for categorical variables.')
            return x_value
        else:
            if x_value >= self.xMax:
                return len(self.objs) - 1
            else:
                return math.floor((x_value-self.xMin)/self.xDelta)

    def get_obj(self, x_value):
        """ :returns the the object in the smallest x break point with x_value greater than or equal.
            In the example above, it returns
                0.1 for 0 <= x_value < 5,
                0.2 for 5 <= x_value < 10, and
                0.3 for x_value >= 10
        """

        return self.objs[self.get_index(x_value)]

    def get_obj_by_index(self, x_index):
        """ :returns the the object in the x break point located at index x_index.
            In the example above, it returns
                0.1 for x_index = 0,
                0.2 for x_index = 1, and
                0.3 for x_index = 2
        """

        return self.objs[x_index]


class OneDimDataFrameWithExpDist (OneDimDataFrame):
    """
    example:
        age,    exponential dist. with mortality rates
        0,      0.1,
        5,      0.2,
        10,     0.3
    """
    def __init__(self, y_values, x_min=0, x_max=1, x_delta=1):
        """
        :param y_values: (list) of rates (in example above: [0.1, 0.2, 0.3]
        :param x_min: minimum value of x (in example above: 0)
        :param x_max: maximum value of x (in example above: 10)
        :param x_delta: interval between break points of x (in example above: 5)
                    if set to 'int', x is treated as categorical variable
        """

        # create the list of exponential distributions
        y_objects = []
        for v in y_values:
            if v <= 0:
                raise ValueError('All y_values (rates of exponential distributions) should be greater than 0.')
            y_objects.append(RVGs.Exponential(scale=1/v))

        OneDimDataFrame.__init__(self, y_objects=y_objects,
                                 x_min=x_min,
                                 x_max=x_max,
                                 x_delta=x_delta)

    def get_dist(self, x_value):
        """ :returns the the exponential distribution in the smallest x break point
            with x_value greater than or equal.
            In the example above, it returns
                exp(0.1) for 0 <= x_value < 5,
                exp(0.2) for 5 <= x_value < 10, and
                exp(0.3) for x_value >= 10
        """
        return self.get_obj(x_value)

    def get_dist_by_index(self, x_index):
        """ :returns the the exponential distribution in the x break point located at index x_index.
            In the example above, it returns
                exp(0.1) for x_index = 0,
                exp(0.2) for x_index = 1, and
                exp(0.3) for x_index = 2
        """
        return self.get_obj_by_index(x_index)


class DataFrameOfObjects:
    """
    example:
        age,   sex,      objects
        0,     0,        obj_0,
        0,     1,        obj_1,
        5,     0,        obj_2,
        5,     1,        obj_3,
        10,    0,        obj_4,
        10,    1,        obj_5
    """

    def __init__(self, list_x_min, list_x_max, list_x_delta, name=''):
        """
        :param list_x_min: list of minimum values of x (in example above: [0, 0] for age and sex)
        :param list_x_max: list of maximum values of x (in example above: [10, 1] for age and sex)
        :param list_x_delta: list of interval between break points of x
                             if set to 'int', x is treated as categorical variable
                             (in example above: [5, 'int'] for age and sex)
        :param name: name of this data frame
        """

        # if x is 1 dimensional, convert min, max, delta into lists
        if not isinstance(list_x_min, list):
            list_x_min = [list_x_min]
        if not isinstance(list_x_max, list):
            list_x_max = [list_x_max]
        if not isinstance(list_x_delta, list):
            list_x_delta = [list_x_delta]

        self._list_x_min = list_x_min
        self._list_x_max = list_x_max
        self._list_x_delta = list_x_delta
        self._xMin = list_x_min[0]
        self._xMax = list_x_max[0]
        self._xDelta = list_x_delta[0]
        self.name = name

        # This class creates (recursively) nested data frames which allow for
        # quick access to the objects given x values.
        # In the example above, it uses the first dimension of x to create a list of data frames.
        # [df_1, df_2, df_3] which correspond to age group [0, 5), [5, 10), and [10+].
        #       df_1 will be data frame over sex with [obj_0, obj_1] as objects.
        #       df_2 will be data frame over sex with [obj_2, obj_3] as objects.
        #       df_2 will be data frame over sex with [obj_4, obj_5] as objects.

        # if this is a 1-dimensional data frame (example above is a 2-dimensional data frame)
        self._ifOneDim = False  # in the example above, False for age category and True for sex category

        # the variables below stores information on the first dimension of x
        # in the example above, it is age.
        self._xs = []               # [0, 5, 10]
        self._ifContinuous = True   # age is a continuous variable

        # find if it is a 1-dimensional data frame
        if len(list_x_min) == 1:
            self._ifOneDim = True
        # find if the first dimensions is a categorical variable
        if list_x_delta[0] == 'int':
            self._ifContinuous = False

        # find x breakpoints of the first dimension
        x = list_x_min[0]
        while x <= list_x_max[0]:
            self._xs.append(x)
            if self._ifContinuous:
                x += list_x_delta[0]
            else:
                x += 1

        # to store y-values:
        # if it is a 1-dimensional data frame, this will contain numbers
        # if it is a multi-dimensional data frame, this will contain data frames that
        # correspond to second, third, etc. dimensions
        self._objs = []
        for x in self._xs:
            if self._ifOneDim:
                self._objs.append(0)
            else:
                self._objs.append(
                    DataFrameOfObjects(list_x_min=list_x_min[1:],
                                       list_x_max=list_x_max[1:],
                                       list_x_delta=list_x_delta[1:]
                                       )
                )

    def __get_index(self, x_value):
        """
        :param x_value: (list) of x values (e.g. value of age and sex in the example above)
        :return: the index of the smallest x break point with x_value[0] greater than or equal.
        """

        # convert to list if x_value is not a list
        if not isinstance(x_value, list):
            x_value = [x_value]

        if not self._ifContinuous:
            if not (isinstance(x_value[0], np.integer) or isinstance(x_value[0], int)):
                raise ValueError('x_value should be an integer for categorical variables.')
            return int(x_value[0] - self._xMin)
        else:
            if x_value[0] > self._xMax:
                return len(self._objs) - 1
            else:
                return int(math.floor((x_value[0] - self._xMin) / self._xDelta))

    def get_sum(self):
        """
        :return: the sum of all objects stored in this data frame (assumes that objects are of type int or float).
        """
        if self._ifOneDim:
            return sum(self._objs)
        else:
            sum_df = 0
            for df in self._objs:
                sum_df += df.get_sum()
            return sum_df

    def get_percents(self):
        """
        :return: (list of lists of lists of ...)
                 Assuming that objects are of type int or float, it returns
                 the object divided by the sum of all objects.
                 in the example above, it returns:
                 [
                    [0, 0, 0.11904761904761904],
                    [0, 1, 0.23809523809523808],
                    [5, 0, 0.13095238095238096],
                    [5, 1, 0.14285714285714285],
                    [10, 0, 0.17857142857142858],
                    [10, 1, 0.19047619047619047]
                ]
        """
        total = self.get_sum()
        rows = []
        for row in self.get_rows():
            a = row[:-1]
            a.extend([row[-1]/total])
            rows.append(a)

        return rows

    def get_index(self, x_value):
        """
        :param x_value: (list) of x values (e.g. value of age and sex in the example above)
        :return: (list) indices that correspond to x values
            in the example above: it returns [0, 1] for 0<= x_value[0]<5 and x_value[1] = 1
        """

        index = self.__get_index(x_value=x_value)
        if self._ifOneDim:
            return [index]
        else:
            return self._objs[index].get_index(x_value=x_value[1:])

    def get_obj(self, x_value):
        """
        :param x_value: (list) of x values (e.g. value of age and sex in the example above)
        :return: the object that correspond to x values
            in the example above: it returns obj_1 for 0<= x_value[0]<5 and x_value[1] = 1
        """

        index = self.__get_index(x_value=x_value)
        if self._ifOneDim:
            return self._objs[index]
        else:
            return self._objs[index].get_obj(x_value[1:])

    def set_obj(self, x_value, obj):
        """
        set the object that correspond to x_value to obj
        :param x_value: (list) of x values (e.g. values of age and sex in the example above)
        :param obj: the new object
        """
        index = self.__get_index(x_value=x_value)
        if self._ifOneDim:
            self._objs[index] = obj
        else:
            self._objs[index].set_obj(x_value=x_value[1:], obj=obj)

    def increment_obj(self, x_value, increment):
        """
        increments the value that correspond to x_value by increment
        :param x_value: (list) of x values (e.g. values of age and sex in the example above)
        :param increment: increment value
        """
        index = self.__get_index(x_value=x_value)
        if self._ifOneDim:
            self._objs[index] += increment
        else:
            self._objs[index].increment_obj(x_value[1:], increment)

    def get_obj_by_index(self, x_index):
        """
        :param x_index: (list) of indices of x values (e.g. index of age and sex in the example above)
        :return: the object that correspond to the indices of x values
            in the example above: it returns obj_1 for x_value [0, 1]
        """
        if self._ifOneDim:
            return self._objs[x_index[0]]
        else:
            return self._objs[x_index[0]].get_obj_by_index(x_index[1:])

    def sample_indices(self, rng):
        """
        :param rng: random number generator
        :return: (list) indices of sampled categories
                        (in the example above, [1, 0] corresponds to age group 1 and sex group 0.
        """

        probs = []
        if self._ifOneDim:
            probs = self._objs
        else:
            for df in self._objs:
                probs.append(df.get_sum())

        empirical_dist = RVGs.Empirical(probabilities=np.array(probs)/sum(probs))

        idx = empirical_dist.sample(rng=rng)

        if self._ifOneDim:
            return [idx]
        else:
            a = [idx]
            a.extend(self._objs[idx].sample_indices(rng))
            return a

    def get_rows(self):
        """
        :return: the data frame in rows. In the example above:
                [
                    [0,     0,        obj_0],
                    [0,     1,        obj_1],
                    [5,     0,        obj_2],
                    [5,     1,        obj_3],
                    [10,    0,        obj_4],
                    [10,    1,        obj_5]
                ]
        """
        rows = []
        if self._ifOneDim:
            for x in self._xs:
                rows.append([x, self.get_obj(x_value=[x])])
        else:
            for i, df in enumerate(self._objs):
                next_rows = df.get_rows()
                for j, row in enumerate(next_rows):
                    a = [self._xs[i]]
                    a.extend(next_rows[j])
                    rows.append(a)

        return rows

    def get_objs(self):
        """
        :return: (list) of objects stored in this data frame
        """
        objs = []
        if self._ifOneDim:
            for x in self._xs:
                objs.append(self.get_obj(x_value=[x]))
        else:
            for df in self._objs:
                objs.extend(df.get_objs())
        return objs

    def get_objs_gen(self):
        """
        :return: a generator function to go over all objects store in this data frame
        """
        if self._ifOneDim:
            for x in self._xs:
                yield self.get_obj(x_value=[x])
        else:
            for df in self._objs:
                for g in df.get_objs_gen():
                    yield g


class DataFrame(DataFrameOfObjects):
    """
    example:
        age,   sex,      mortality rate
        0,     0,        0.1,
        0,     1,        0.11,
        5,     0,        0.2,
        5,     1,        0.21,
        10,    0,        0.3
        10,    1,        0.31
    """
    def __init__(self, rows, list_x_min, list_x_max, list_x_delta):
        """
        :param rows: (list of list) the table above
        :param list_x_min: list of minimum value of x (in example above: [0, 0])
        :param list_x_max: list of maximum value of x (in example above: [10, 1])
        :param list_x_delta: list of interval between break points of x
                    if set to 'int', x is treated as categorical variable
                    (in example above: [5, 'int'])
        """

        DataFrameOfObjects.__init__(self,
                                    list_x_min=list_x_min,
                                    list_x_max=list_x_max,
                                    list_x_delta=list_x_delta)

        for row in rows:
            self.set_obj(x_value=row[0:-1], obj=row[-1])

    def get_value_by_index(self, x_index):

        return self.get_obj_by_index(x_index)

    def get_value(self, x_value):

        return self.get_obj(x_value)


class DataFrameWithExpDist(DataFrameOfObjects):
    """
    example:
        age,   sex,      mortality rate
        0,     0,        0.1,
        0,     1,        0.11,
        5,     0,        0.2,
        5,     1,        0.21,
        10,    0,        0.3
        10,    1,        0.31
    """
    def __init__(self, rows, list_x_min, list_x_max, list_x_delta):
        """
        :param rows: (list of list) the table above
        :param list_x_min: list of minimum value of x (in example above: [0, 0])
        :param list_x_max: list of maximum value of x (in example above: [10, 1])
        :param list_x_delta: list of interval between break points of x
                    if set to 'int', x is treated as categorical variable
                    (in example above: [5, 'int'])
        """

        DataFrameOfObjects.__init__(self,
                                    list_x_min=list_x_min,
                                    list_x_max=list_x_max,
                                    list_x_delta=list_x_delta)

        for row in rows:
            self.set_obj(x_value=row[0:-1],
                         obj=RVGs.Exponential(scale=1/row[-1]))

    def get_dist(self, x_value):

        return self.get_obj(x_value)


class DataFrameWithEmpiricalDist(DataFrame):
    """
    example:
        age,   sex,      probability
        0,     0,        0.1,
        0,     1,        0.2,
        5,     0,        0.3,
        5,     1,        0.4,
    """
    def __init__(self, rows, list_x_min, list_x_max, list_x_delta):
        """
        :param rows: (list of list) the table above
        :param list_x_min: list of minimum value of x (in example above: [0, 0])
        :param list_x_max: list of maximum value of x (in example above: [10, 1])
        :param list_x_delta: list of interval between break points of x
                    if set to 'int', x is treated as categorical variable
                    (in example above: [5, 'int'])
        """

        DataFrame.__init__(self, rows=rows,
                           list_x_min=list_x_min,
                           list_x_max=list_x_max,
                           list_x_delta=list_x_delta)
        self.listXDelta = list_x_delta

        # make sure probabilities add to 1
        sum_probs = 0
        for row in rows:
            sum_probs += row[-1]
        if sum_probs < 0.99999 or sum_probs > 1.000001:
            raise ValueError('Sum of probabilities should add to 1.')

    def sample_values(self, rng):
        """
        :param rng: random number generator
        :return: (list) values of categories
                (in the example above, [1.4, 0] corresponds to age 1.4 and sex group 0.
        """

        values = []
        idx = self.sample_indices(rng=rng)
        for i, deltaX in enumerate(self.listXDelta):
            if deltaX == 'int':
                values.append(idx[i]+self._list_x_min[i])
            else:
                unif_dist = RVGs.Uniform(loc=0, scale=deltaX)
                values.append(idx[i]*deltaX + self._list_x_min[i] + unif_dist.sample(rng=rng))

        return values


class Pyramid(DataFrameOfObjects):
    """
    example:
        age,   sex,      Prevalence
        0,     0,        10,
        0,     1,        20,
        5,     0,        30,
        5,     1,        40,
        10,    0,        50,
        10,    1,        60
    """

    def __init__(self, list_x_min, list_x_max, list_x_delta, name=''):
        """
        :param list_x_min: list of minimum value of x (in example above: [0, 0])
        :param list_x_max: list of maximum value of x (in example above: [10, 1])
        :param list_x_delta: list of interval between break points of x
                    if set to 'int', x is treated as categorical variable
                    (in example above: [5, 'int'])
        """

        DataFrameOfObjects.__init__(self,
                                    list_x_min=list_x_min,
                                    list_x_max=list_x_max,
                                    list_x_delta=list_x_delta,
                                    name=name)

    def record_increment(self, x_values, increment):
        """
        updates the value of a group in the pyramid
        :param x_values: (list) group to be updated
                (e.g. [1.2, 0] corresponds to age 1.2 and sex 0 in the example above)
        :param increment: (integer) change (+ or -) in the value of the pyramid at the specified group
        """

        self.increment_obj(x_value=x_values, increment=increment)

    def record_value(self, x_values, value):
        """
        updates the value of a group in the pyramid
        :param x_values: (list) group to be updated
                (e.g. [1.2, 0] corresponds to age 1.2 and sex 0 in the example above)
        :param value: value of the pyramid at the specified group
        """

        self.set_obj(x_value=x_values, obj=value)

    def record_values_from_another_pyramid(self, another_pyramid):
        """ updates this pyramid using another pyramid """

        for row in another_pyramid.get_rows():
            self.record_value(x_values=row[:-1],
                              value=row[-1])

    def get_current_value(self, x_values):
        """
        :returns the value of the pyramid at the specified group
        :param x_values: (list) group
                (e.g. [1.2, 0] corresponds to age 1.2 and sex 0 in the example above)
        """

        return self.get_obj(x_value=x_values)

    def get_values(self):
        """ :returns the value of the pyramid in the table format
            In the example above, it returns:
            [
            [0,     0,        10],
            [0,     1,        20],
            [5,     0,        30],
            [5,     1,        40],
            [10,    0,        50],
            [10,    1,        60]
            ]

        """
        return self.get_rows()

    def get_percentages(self):
        """ :returns the percentage of the population in each group
            In the example above, it returns:
            [
            [0,     0,        10/210],
            [0,     1,        20/210],
            [5,     0,        30/210],
            [5,     1,        40/210],
            [10,    0,        50/210],
            [10,    1,        60/210]
            ]

        """
        return self.get_percents()



import heapq


class SimulationEvent:
    def __init__(self, time, priority):
        """
        :param time: (float) time of the event
        :param priority: priority of the event (the lowest value implies the highest priority)
        """
        self.time = time            # event time
        self.priority = priority    # event priority

    def process(self, rng=None):
        """ implements instruction to process this event once occurs
        abstract method to be overridden in derived classes to process an event
        :param rng: random number generator
        """

        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")


class SimulationCalendar:
    def __init__(self):
        """  create a simulation calendar """

        self._q = []  # a list (priority queue) to store simulation events
        self.time = 0   # current time

    def n_events(self):
        """
        :return: number of scheduled events"""

        return len(self._q)

    def add_event(self, event):
        """ add a new event to the calendar (sorted by event time and then priority)
        :param event: a simulation event to be added to the simulation calendar """

        if event.time < self.time:
            raise ValueError('An event with event time less than the current time cannot be added to the calendar.')

        entry = [event.time, event.priority, event]
        heapq.heappush(self._q, entry)

    def get_next_event(self):
        """
        :return: the next simulation event
        (if time of events are equal, the event with the lowest value for priority will be returned.) """

        self.time, priority, next_event = heapq.heappop(self._q)
        return next_event

    def clear_calendar(self):
        """ deletes all scheduled events but keeps the current time """

        self._q.clear()

    def reset(self):
        """ deletes all scheduled events and resets the current time to zero """

        self.time = 0
        self._q.clear()



import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d

from deampy.plots.plot_support import output_figure, calculate_ticks, format_axis_tick_labels


def format_ax(ax,
              x_range=None, x_delta=None,
              y_range=None, y_delta=None, if_y_axis_prob=False,
              x_axis_format_deci=None, y_axis_format_deci=None):

    # the range of x and y-axis are set so that we can get the
    # tick values and label
    if y_range is None and if_y_axis_prob:
        ax.set_ylim((-0.01, 1.01))
    if y_range:
        ax.set_ylim(y_range)
    if x_range:
        ax.set_xlim(x_range)

    # get x ticks
    if x_delta is None:
        vals_x = ax.get_xticks()
    else:
        vals_x = calculate_ticks(interval=x_range, delta=x_delta)

    # get y ticks
    if y_delta is None:
        vals_y = ax.get_yticks()
    else:
        vals_y = calculate_ticks(interval=y_range, delta=y_delta)

    # format x-axis
    ax.set_xticks(vals_x)
    if x_axis_format_deci is None:
        format_axis_tick_labels(ax=ax, axis='x', format_deci=(',', 0))
    else:
        format_axis_tick_labels(ax=ax, axis='x', format_deci=x_axis_format_deci)

    # d = 2 * (x_range[1] - x_range[0]) / 200
    # ax.set_xlim([x_range[0] - d, x_range[1] + d])

    # format y-axis
    # if y_range is None:
    ax.set_yticks(vals_y)
    if if_y_axis_prob:
        format_axis_tick_labels(ax=ax, axis='y', format_deci=(None, 1))
    if y_axis_format_deci is not None:
        format_axis_tick_labels(ax=ax, axis='y', format_deci=y_axis_format_deci)

    if y_range is None and if_y_axis_prob:
        ax.set_ylim((-0.01, 1.01))
    elif y_range:
        ax.set_ylim(y_range)
    else:
        ax.set_ylim([vals_y[0], vals_y[-1]])

    if not if_y_axis_prob:
        ax.axhline(y=0, c='k', ls='--', linewidth=0.5)


def add_grids(ax, grid_info):

    # grid
    if grid_info is None:
        pass
    elif grid_info == 'default':
        color, linestyle, linewidth, alpha = ('k', '--', 0.5, 0.2)
    else:
        color, linestyle, linewidth, alpha = grid_info
    if grid_info is not None:
        ax.grid(color=color, linestyle=linestyle, linewidth=linewidth, alpha=alpha)


def add_curves_to_ax(ax, curves, title=None,
                     x_range=None, x_label=None, y_label=None, y_range=None,
                     y_axis_multiplier=1,
                     x_delta=None,
                     transparency_lines=0.5,
                     transparency_intervals=0.2,
                     legends=None,
                     show_legend=False,
                     show_frontier=True,
                     show_labels_on_frontier=False,
                     curve_line_width=1.0, frontier_line_width=4.0,
                     if_y_axis_prob=False, y_axis_format_decimals=None,
                     legend_font_size_and_loc=(7, 'upper left'),
                     frontier_label_shift_x=-0.01,
                     frontier_label_shift_y=0.01,
                     grid_info=None):

    for i, curve in enumerate(curves):

        label = curve.label if legends is None else legends

        # if to interpolate points to create a smooth curve
        interpolate = False
        if interpolate:
            f2 = interp1d(curve.xs, curve.ys * y_axis_multiplier, kind='cubic')
            x_smooth = np.linspace(curve.xs.min(), curve.xs.max(), 100)
            y_smooth = f2(x_smooth)
            xs = x_smooth
            ys = y_smooth
        else:
            xs = curve.xs
            ys = curve.ys * y_axis_multiplier

        # plot line
        ax.plot(xs, ys,
                c=curve.color, alpha=transparency_lines,
                linewidth=curve_line_width, linestyle=curve.linestyle, label=label)

        # plot intervals
        if curve.l_errs is not None and curve.u_errs is not None:
            ax.fill_between(curve.xs,
                            (curve.ys - curve.l_errs) * y_axis_multiplier,
                            (curve.ys + curve.u_errs) * y_axis_multiplier,
                            color=curve.color, alpha=transparency_intervals)
        # plot frontier
        if show_frontier:
            # check if this strategy is not dominated
            if curve.frontierXs is not None and len(curve.frontierXs) > 0:
                y = [y*y_axis_multiplier if y is not None else None for y in curve.frontierYs]
                ax.plot(curve.frontierXs, y,
                        c=curve.color, alpha=1, linewidth=frontier_line_width)

    if show_legend:
        ax.legend(fontsize=legend_font_size_and_loc[0], loc=legend_font_size_and_loc[1]) #loc=2,

    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_ylim(y_range)

    # add labels on the frontier
    if show_labels_on_frontier:
        y_min, y_max = ax.get_ylim()
        y_axis_length = y_max - y_min
        for curve in curves:
            if curve.frontierXs is not None and len(curve.frontierXs) > 0:
                if curve.frontierYs[0] is not None and curve.frontierYs[-1] is not None:
                    x_axis_length = x_range[1] - x_range[0]
                    x = 0.5 * (curve.frontierXs[0] + curve.frontierXs[-1]) + frontier_label_shift_x * x_axis_length
                    y = 0.5 * (curve.frontierYs[0] + curve.frontierYs[-1]) * y_axis_multiplier \
                        + frontier_label_shift_y * y_axis_length
                    ax.text(x=x, y=y, s=curve.label, fontsize=legend_font_size_and_loc[0], c=curve.color)

    # grids
    add_grids(ax=ax, grid_info=grid_info)

    # do the other formatting
    format_ax(ax=ax, y_range=y_range,
              x_range=x_range, x_delta=x_delta,
              if_y_axis_prob=if_y_axis_prob,
              y_axis_format_deci=y_axis_format_decimals)


def add_min_monte_carlo_samples_to_ax(
        ax, dict_of_ns, epsilons, x_range=None, y_range=None, x_multiplier=1):

    colors = ('purple', 'blue', 'green', 'red')
    markers = ('o', 'v', '^', 's')

    adjusted_epsilons = [x*x_multiplier for x in epsilons]

    i = 0
    for alpha in dict_of_ns:
        ns = [dict_of_ns[alpha][key] for key in dict_of_ns[alpha]]

        # build n values and intervals
        n_values = []
        l_errs = []
        u_errs = []
        for n in ns:
            if isinstance(n, tuple):
                value, interval = n
                n_values.append(value)
                l_errs.append(value - interval[0])
                u_errs.append(interval[1] - value)
            else:
                n_values.append(n)

        ax.scatter(adjusted_epsilons, n_values, marker=markers[i], color=colors[i],
                   label=r'$\alpha=${:.{prec}%}'.format(alpha, prec=0))
        # error bars
        if len(l_errs) > 0:
            ax.errorbar(
                adjusted_epsilons, n_values, yerr=[l_errs, u_errs], fmt='none',
                ecolor=colors[i], capsize=3, linewidth=0.75)

        ax.plot(adjusted_epsilons, n_values, 'k--', color=colors[i], linewidth=0.5)
        i += 1

    if y_range:
        ax.set_ylim(y_range)
    else:
        ax.set_ylim(0)

    if x_range:
        ax.set_xlim(x_range)

    ax.set_xticks(adjusted_epsilons)
    vals_x = ax.get_xticks()
    ax.set_xticklabels(['${:,.{prec}f}'.format(x, prec=0) for x in vals_x])

    vals_y = ax.get_yticks()
    ax.set_yticklabels(['{:,.{prec}f}'.format(y, prec=0) for y in vals_y])

    ax.legend(fontsize=8)


def add_icer_over_itr_to_ax(ax, icer_over_itr, x_range=None, y_range=None, y_delta=None):

    ax.plot(icer_over_itr, 'b-', linewidth=0.5)

    # if y_range:
    #     ax.set_ylim(y_range)
    #
    # if x_range:
    #     ax.set_xlim(x_range)

    format_ax(ax=ax,
              x_range=x_range, x_axis_format_deci=(',', 0),
              y_range=y_range, y_delta=y_delta, y_axis_format_deci=('$', 0))
    #
    #
    # format_axis_tick_labels(ax=ax, axis='x', format_deci=(',', 0))
    #
    #
    # if y_delta is None:
    #     vals_y = ax.get_yticks()
    # else:
    #     vals_y = calculate_ticks(interval=y_range, delta=y_delta)
    # ax.set_yticks(vals_y)
    # ax.set_yticklabels(['${:,.0f}'.format(y) for y in vals_y])


def plot_icer_over_itrs(icer_over_itr, x_range=None, y_range=None, legend=None, figure_size=None, file_name=None):

    fig, ax = plt.subplots(figsize=figure_size)

    # add icer over iterations
    add_icer_over_itr_to_ax(ax=ax, icer_over_itr=icer_over_itr, x_range=x_range, y_range=y_range)

    # add legend if provided
    if legend is not None:
        ax.legend([legend])

    # output figure
    output_figure(fig, file_name)



from scipy.stats import spearmanr, pearsonr, rankdata

import deampy.format_functions as F
import deampy.in_out_functions as IO
from deampy.statistics import partial_corr


class SensitivityAnalysis:
    def __init__(self, dic_parameter_values, output_values):
        """
        :param dic_parameter_values: (dictionary) of parameter values with parameter names as the key
        :param output_values: (list) of output values (e.g. cost or QALY observations)
        """

        self.dicParameterValues = dic_parameter_values
        self.outputValues = output_values

        for paramName, paramValues in dic_parameter_values.items():

            if len(paramValues) != len(output_values):
                raise ValueError('Number of parameter values should be equal to the number of output values. '
                                 'Error in parameter "{}".'.format(paramName))

    def get_pearson_corr(self):
        """ :returns (list) of [parameter name, Pearson's correlation coefficients, p-value] """

        return self._get_corr(f=pearsonr)

    def get_spearman_corr(self):
        """ :returns (list) of [parameter name, Spearman's rank correlation coefficients, p-value] """

        return self._get_corr(f=spearmanr)

    def get_partial_corr(self):
        """ :returns (list) of [parameter name, partial correlation coefficients, p-value] """

        return self._get_partial_corr(dic_param_values=self.dicParameterValues, output_values=self.outputValues)

    def get_partial_rank_corr(self):
        """ :returns (list) of [parameter name, partial rank correlation coefficients, p-value] """

        # find ranked output values
        ranked_outputs = rankdata(self.outputValues)
        # find ranked parameter values
        dic_ranked_param_values = dict()
        for paramName, paramValues in self.dicParameterValues.items():
            dic_ranked_param_values[paramName] = rankdata(paramValues)

        return self._get_partial_corr(dic_param_values=dic_ranked_param_values, output_values=ranked_outputs)

    def print_corr(self, corr='r'):
        """
        :param corr: 'r' for Pearson's,
                     'rho' for Spearman's,
                     'p' for partial correlation, and
                     'pr' for partial rank correlation
        :return:
        """

        results, text = self._get_results_text(corr=corr)
        print(text)
        for par, values in results.items():
            print('\t', par, '{} , {}'.format(values[0], values[1]))

    def export_to_csv(self, corrs='r', file_name='Correlations.csv', decimal=3, delimiter=','):
        """
        formats the correlation coefficients and p-value to the specified decimal point and exports to a csv file
        :param corrs: (string) or (list of strings) from
                     'r' for Pearson's,
                     'rho' for Spearman's rank correlation,
                     'p' for partial correlation, and
                     'pr' for partial rank correlation
        :param file_name: file name
        :param decimal: decimal points to round the estimates to
        :param delimiter: to separate by comma, use ',' and by tab, use '\t'
        """

        if not isinstance(corrs, list):
            corrs = [corrs]

        # make the header
        tile_row = ['Parameter']
        for corr in corrs:
            tile_row.append(self._full_name(corr=corr))
            tile_row.append('P-value')

        # add the header
        formatted_results = [tile_row]

        # add the names of parameters (first column)
        for name in self.dicParameterValues:
            formatted_results.append([name])

        # calculate all forms of correlation requested
        for corr in corrs:
            results, text = self._get_results_text(corr=corr)

            i = 1
            for par, values in results.items():
                coef = F.format_number(number=values[0], deci=decimal)
                p = F.format_number(number=values[1], deci=decimal)
                formatted_results[i].extend([coef, p])
                i += 1

        IO.write_csv(file_name=file_name, rows=formatted_results, delimiter=delimiter)

    def _get_results_text(self, corr):

        if corr == 'r':
            text = "Pearson's correlation coefficients and p-values"
            results = self.get_pearson_corr()
        elif corr == 'rho':
            text = "Spearman's rank correlation coefficients and p-values"
            results = self.get_spearman_corr()
        elif corr == 'p':
            text = "Partial correlation coefficients and p-values"
            results = self.get_partial_corr()
        elif corr == 'pr':
            text = "Partial rank correlation coefficients and p-values"
            results = self.get_partial_rank_corr()
        else:
            raise ValueError('Invalid correlation type is provided.')

        return results, text

    @staticmethod
    def _full_name(corr):

        if corr == 'r':
            return "Pearson's correlation"
        elif corr == 'rho':
            return "Spearman's correlation"
        elif corr == 'p':
            return 'Partial correlation'
        elif corr == 'pr':
            return 'Partial rank correlation'
        else:
            raise ValueError('Invalid correlation type is provided.')

    def _get_corr(self, f):
        """
        f: correlation function
        :returns (list) of [parameter name, correlation coefficients, p-value] """

        result = dict()  # each row [parameter name, correlation, p-value]
        for paramName, paramValues in self.dicParameterValues.items():
            # calculate Spearman rank-order correlation coefficient
            coef, p = f(paramValues, self.outputValues)
            # store [parameter name, Spearman coefficient, and p-value
            result[paramName] = [coef, p]

        return result

    @staticmethod
    def _get_partial_corr(dic_param_values, output_values):

        result = dict()  # each row [parameter name, correlation, p-value]
        for paramName, paramValues in dic_param_values.items():

            z = []
            for otherParamName, otherParamValues in dic_param_values.items():
                if paramName != otherParamName:
                    z.append(otherParamValues)

            # calculate partial correlation coefficient
            coef, p = partial_corr(x=paramValues, y=output_values, z=z)
            # store [parameter name, Spearman coefficient, and p-value
            result[paramName] = [coef, p]

        return result



from pathlib import Path

from setuptools import setup

# read the contents of your README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='deampy',
    version='1.5.6',
    install_requires=['numpy', 'numba', 'matplotlib', 'scipy', 'statsmodels', 'scikit-learn', 'pandas', 'seaborn'],
    packages=['deampy', 'deampy.optimization', 'deampy.plots', 'deampy.support'],
    url='https://github.com/modeling-health-care-decisions/deampy',
    license='MIT License',
    author='Reza Yaesoubi',
    author_email='reza.yaesoubi@yale.edu',
    description='Decision analysis in medicine and public health',
    long_description=long_description
)



import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

from deampy.plots.plot_support import output_figure
from deampy.sample_path import *


def plot_sample_path(sample_path,
                     title=None, x_label=None, y_label=None,
                     x_range=None, y_range=None,
                     figure_size=None, file_name=None,
                     legend=None, color=None, transparency=0.75, connect='step'):
    """
    plot a sample path
    :param sample_path: a sample path
    :param title: (string) title of the figure
    :param x_label: (string) x-axis label
    :param y_label: (string) y-axis label
    :param x_range: (list) [x_min, x_max]
    :param y_range: (list) [y_min, y_max]
    :param figure_size: (tuple) figure size
    :param file_name: (string) filename to save the histogram as (e.g. 'fig.png')
    :param legend: (string) the legend
    :param color: (string) for example: 'b' blue 'g' green 'r' red 'c' cyan 'm' magenta 'y' yellow 'k' black
    :param transparency: (float) between 0 and 1 for the transparency of curves
    :param connect: (string) set to 'step' to produce a step graph and to 'line' to produce a line graph
    """

    if not (isinstance(sample_path, IncidenceSamplePath) or isinstance(sample_path, PrevalenceSamplePath)):
        raise ValueError(
            'sample_path should be an instance of PrevalenceSamplePath or PrevalencePathBatchUpdate.')

    fig, ax = plt.subplots(figsize=figure_size)

    # add a sample path to this ax
    add_sample_path_to_ax(sample_path=sample_path,
                          ax=ax,
                          color=color, transparency=transparency,
                          legend=legend,
                          connect=connect,
                          x_label=x_label, y_label=y_label, title=title)
    ax.set_ylim(bottom=0)  # the minimum has to be set after plotting the values

    if x_range is not None:
        ax.set_xlim(x_range)
    else:
        if isinstance(sample_path, PrevalenceSamplePath):
            ax.set_xlim(left=0)
        elif isinstance(sample_path, IncidenceSamplePath):
            ax.set_xlim(left=0.5)

    if y_range is not None:
        ax.set_ylim(y_range)

    # output figure
    output_figure(fig, file_name)


def plot_sample_paths(sample_paths,
                      title=None, x_label=None, y_label=None,
                      x_range=None, y_range=None,
                      figure_size=None, file_name=None,
                      legends=None, legend_fontsize=8, transparency=1,
                      color_codes=None, delete_initial_zeroes=False,
                      common_color_code=None, connect='step',
                      obs_info=None):
    """ graphs multiple sample paths
    :param sample_paths: a list of sample paths
    :param title: (string) title of the figure
    :param x_label: (string) x-axis label
    :param y_label: (string) y-axis label
    :param x_range: (list) [x_min, x_max]
    :param y_range: (list) [y_min, y_max]
    :param figure_size: (tuple) figure size
    :param file_name: (string) filename to save the histogram as (e.g. 'fig.png')
    :param legends: (list) of strings for legend
    :param legend_fontsize: (float) legend font size
    :param transparency: (float) 0.0 transparent through 1.0 opaque
    :param color_codes: (list) of color code for sample paths
    :param common_color_code: (string) color code if all sample paths should have the same color
        'b'	blue 'g' green 'r' red 'c' cyan 'm' magenta 'y' yellow 'k' black
    :param connect: (string) set to 'step' to produce an step graph and to 'line' to produce a line graph
    :param obs_info: (dict) with keys 'ts' and 'obss' for observed values
    """

    if len(sample_paths) == 1:
        raise ValueError('Only one sample path is provided. Use graph_sample_path instead.')

    fig, ax = plt.subplots(figsize=figure_size)
    ax.set_title(title)  # title
    ax.set_xlabel(x_label)  # x-axis label
    ax.set_ylabel(y_label)  # y-axis label
    if x_range is not None:
        ax.set_xlim(x_range)
    if y_range is not None:
        ax.set_ylim(y_range)

    # add all sample paths
    add_sample_paths_to_ax(sample_paths=sample_paths,
                           ax=ax,
                           color_codes=color_codes,
                           common_color_code=common_color_code,
                           transparency=transparency,
                           connect=connect,
                           delete_initial_zeroes=delete_initial_zeroes)

    if obs_info is not None:
        add_obs_to_ax(
            ax=ax,
            ts=obs_info['observed_times'],
            obss=obs_info['observed_values'],
            color=obs_info['observed_color_code'],
            label=obs_info['observed_label'])

    # add legend if provided
    if legends is not None:
        if common_color_code is None:
            ax.legend(legends, fontsize=legend_fontsize)
        else:
            ax.legend([legends], fontsize=legend_fontsize)

        # this is to make sure that the legend transparency is set to 1
        if obs_info is not None:
            legend_lines = [
                Line2D([0], [0], color=common_color_code, alpha=1),
                Line2D([0], [0],
                       color=obs_info['observed_color_code'],
                       linestyle='-', marker='.', markersize=10, alpha=1)
            ]
            ax.legend(handles=legend_lines, labels=legends, fontsize=8)

    # set the minimum of y-axis to zero
    if y_range is None:
        ax.set_ylim(bottom=0)  # the minimum has to be set after plotting the values

    # output figure
    output_figure(fig, file_name)


def plot_sets_of_sample_paths(sets_of_sample_paths,
                              title=None, x_label=None, y_label=None,
                              x_range=None, y_range=None,
                              figure_size=None, file_name=None,
                              legends=None, transparency=1, color_codes=None, connect='step'):
    """ graphs multiple sample paths
    :param sets_of_sample_paths: (list) of list of sample paths
    :param title: (string) title of the figure
    :param x_label: (string) x-axis label
    :param y_label: (string) y-axis label
    :param x_range: (list) [x_min, x_max]
    :param y_range: (list) [y_min, y_max]
    :param figure_size: (tuple) figure size
    :param file_name: (string) filename to to save the histogram as (e.g. 'fig.png')
    :param legends: (list of strings) for legends
    :param transparency: float (0.0 transparent through 1.0 opaque)
    :param color_codes: (list of strings) color code of sample path sets
            e.g. 'b' blue 'g' green 'r' red 'c' cyan 'm' magenta 'y' yellow 'k' black
    :param connect: (string) set to 'step' to produce an step graph and to 'line' to produce a line graph
    """

    if len(sets_of_sample_paths) == 1:
        raise ValueError('Only one set of sample paths is provided. Use plot_sample_paths instead.')

    fig, ax = plt.subplots(figsize=figure_size)
    if x_range is not None:
        ax.set_xlim(x_range)
    if y_range is not None:
        ax.set_ylim(y_range)

    # add all sample paths
    add_sets_of_sample_paths_to_ax(sets_of_sample_paths=sets_of_sample_paths,
                                   ax=ax,
                                   color_codes=color_codes,
                                   legends=legends,
                                   transparency=transparency,
                                   connect=connect)

    ax.set_title(title)  # title
    ax.set_xlabel(x_label)  # x-axis label
    ax.set_ylabel(y_label)  # y-axis label
    # # set the minimum of y-axis to zero
    # ax.set_ylim(bottom=0)  # the minimum has to be set after plotting the values
    # output figure
    output_figure(fig, file_name)


def add_sample_path_to_ax(sample_path, ax, legend=None, color=None, transparency=1.0, connect='step',
                          title=None, x_label=None, y_label=None,
                          x_range=None, y_range=None, legend_fontsize=8,
                          delete_initial_zeroes = False
                          ):

    # x and y values
    if isinstance(sample_path, PrevalenceSamplePath):
        x_values = sample_path.get_times()
    elif isinstance(sample_path, IncidenceSamplePath):
        x_values = sample_path.get_period_numbers()

    y_values = sample_path.get_values()
    x_values = x_values[:len(y_values)]  # make sure that x and y values have the same length

    # plot the sample path
    if len(y_values) > 0:
        if connect == 'step':
            ax.step(x=x_values, y=y_values, where='post', color=color,
                    linewidth=0.75, label=legend, alpha=transparency)
        else:
            ax.plot(x_values, y_values, color=color,
                    linewidth=0.75, label=legend, alpha=transparency)

    if title is not None:
        ax.set_title(title)
    if x_label is not None:
        ax.set_xlabel(x_label)
    if y_label is not None:
        ax.set_ylabel(y_label)
    ax.set_xlim(x_range)
    ax.set_ylim(y_range)

    # add legend if provided
    if legend is not None:
        ax.legend(fontsize=legend_fontsize)


def add_sample_paths_to_ax(sample_paths, ax, color_codes=None, common_color_code=None, transparency=0.5,
                           connect='step', legends=None, legend_fontsize=8,
                           title=None, x_label=None, y_label=None,
                           x_range=None, y_range=None, delete_initial_zeroes=False):

    # add every path
    for i, path in enumerate(sample_paths):
        if color_codes is not None:
            color = color_codes[i]
        elif common_color_code is not None:
            color = common_color_code
        else:
            color = None

        legend = None if legends is None else legends[i]

        add_sample_path_to_ax(sample_path=path,
                              ax=ax,
                              color=color,
                              legend=legend,
                              legend_fontsize=legend_fontsize,
                              transparency=transparency,
                              connect=connect,
                              delete_initial_zeroes=delete_initial_zeroes)

    if title is not None:
        ax.set_title(title)
    if x_label is not None:
        ax.set_xlabel(x_label)
    if y_label is not None:
        ax.set_ylabel(y_label)
    if x_range is not None:
        ax.set_xlim(x_range)
    if y_range is not None:
        ax.set_ylim(y_range)


def add_sets_of_sample_paths_to_ax(
        sets_of_sample_paths, ax, color_codes, legends, transparency, connect='step',
        title=None, x_label=None, y_label=None,
        x_range=None, y_range=None):

    # add every path
    for i, sample_paths in enumerate(sets_of_sample_paths):
        for j, path in enumerate(sample_paths):
            if j == 0:
                legend = legends[i]
            else:
                legend = None
            if color_codes is None:
                this_color_code = None
            else:
                this_color_code = color_codes[i]
            add_sample_path_to_ax(sample_path=path,
                                  ax=ax,
                                  color=this_color_code,
                                  legend=legend,
                                  transparency=transparency,
                                  connect=connect)

    if title is not None:
        ax.set_title(title)
    if x_label is not None:
        ax.set_xlabel(x_label)
    if y_label is not None:
        ax.set_ylabel(y_label)
    ax.set_xlim(x_range)
    ax.set_ylim(y_range)

    # this is to make sure that the legend transparency is set to 1
    legend_lines = [
        Line2D([0], [0], color=color_codes[i], alpha=1, label=legends[i]) for i in range(len(legends))
    ]
    ax.legend(handles=legend_lines, fontsize=8)


def add_obs_to_ax(ax, ts, obss, color='black', label='Observed', marker_size=25, linewidth=1, alpha=1):
    """ add observations to the ax
    :param ax: the axis to add the observations to
    :param obss: a list of observations
    """
    ax.scatter(
        x=ts, y=obss, color=color,
        label=label, alpha=alpha, s=marker_size)
    ax.plot(
        ts, obss, color=color, linestyle='-', marker='.',
        alpha=alpha, linewidth=linewidth, label=label
    )



import random

import numpy as np
from scipy.stats import norm


def proper_file_name(text):
    """
    :param text: file name
    :return: filename where invalid characters are removed
    """

    return text.replace('|', ',').replace('<', 'l').replace('>', 'g').replace('\n', '')


def get_cumulative_mean(data):
    """
    calculates the cumulative mean of a time-series
    :param data: list of observations
    :return: list of cumulative means
    """

    cum_means = []
    total = 0
    n = 0
    for i in range(len(data)):
        if data[i] is not None:
            total += data[i]
            n += 1
        if n > 0:
            cum_means.append(total/n)
        else:
            cum_means.append(None)

    return cum_means


def get_cumulative_var(data):
    """
    calculates the cumulative variance of a time-series
    :param data: list of observations
    :return: list of cumulative variances
    """

    cum_vars = []
    total = 0
    total_sq = 0
    n = 0
    for i in range(len(data)):
        if data[i] is not None:
            total += data[i]
            total_sq += data[i]**2
            n += 1
        if n > 1:
            cum_vars.append((total_sq - (total**2)/n)/(n-1))
        else:
            cum_vars.append(None)

    return cum_vars


def get_moving_average(data, window=2):
    """
    calculates the moving average of a time-series
    :param data: list of observations
    :param window: the window (number of data points) over which the average should be calculated
    :return: list of moving averages
    """

    if window < 2:
        raise ValueError('The window over which the moving averages '
                         'should be calculated should be greater than 1.')
    if window >= len(data):
        window = len(data) - 1
        # raise ValueError('The window over which the moving averages '
        #                  'should be calculated should be less than the number of data points.')

    window = int(window)

    averages = []

    # the 'window - 1' averages cannot be calculated
    for i in range(window-1):
        averages.append(None)

    # calculate the first moving average
    total = 0
    n = 0
    for i in range(window):
        if data[i] is not None:
            total += data[i]
            n += 1
    moving_ave = total/n
    averages.append(moving_ave)

    for i in range(window, len(data)):
        if moving_ave is None:
            moving_sum = 0
        else:
            moving_sum = moving_ave * n
        if data[i-window] is not None:
            moving_sum -= data[i-window]
            n -= 1
        if data[i] is not None:
            moving_sum += data[i]
            n += 1

        if n > 0:
            moving_ave = moving_sum/n
        else:
            moving_ave = None
        averages.append(moving_ave)

    return averages


def effective_sample_size(likelihood_weights):
    """
    :param likelihood_weights: (list) probabilities
    :returns: the effective sample size
    """

    # convert to numpy array if needed
    if not type(likelihood_weights) == np.ndarray:
        likelihood_weights = np.array(likelihood_weights)

    # normalize the weights if needed
    s = sum(likelihood_weights)
    if s < 0.99999 or s > 1.00001:
        likelihood_weights = likelihood_weights / s

    return 1 / np.sum(likelihood_weights ** 2)


def get_random_colors(n, seed=0):
    """
    :param n: (int) number of colors randomly selected 
    :param seed: (int) seed of the random number generator 
    :return: (list) of color codes 
    """
    
    random.seed(seed)

    colors = []
    for j in range(n):
        rand_colors = "#" + ''.join([random.choice('ABCDEF0123456789') for i in range(6)])
        colors.append(rand_colors)
    
    return colors


def convert_lnl_to_prob(ln_likelihoods):

    for i, lnl in enumerate(ln_likelihoods):
        if np.isnan(lnl):
            ln_likelihoods[i] = -np.inf

    # find the maximum lnl
    max_lnl = max(ln_likelihoods)

    # find probability weights
    probs = [np.exp(s - max_lnl) for s in ln_likelihoods]
    sum_prob = sum(probs)

    # normalize the probability weights
    return np.array(probs)/sum_prob


def get_prob_normal_greater_0(mean, st_dev):
    """
    :param mean: (float) mean of the normal distribution
    :param st_dev: (float) standard deviation of the normal distribution
    :return: the probability of the random variable following this normal distribution is greater than 0
    """

    return 1 - norm(loc=mean, scale=st_dev).cdf(0)


def get_prob_x_greater_than_y(x_mean, x_st_dev, y_mean, y_st_dev, corr=0):
    """
    :param x_mean: (float) the mean of x, which follows a normal distribution.
    :param x_st_dev: (float) the standard deviation of x
    :param y_mean: (float)the mean of y, which follows a normal distribution
    :param y_st_dev: (float) the standard deviation of y
    :param corr: (float) the correlation between x and y
    :return: the probability that x > y
    """

    return get_prob_normal_greater_0(
        mean=x_mean - y_mean, st_dev=np.sqrt(x_st_dev**2 + y_st_dev**2 + 2*corr*x_st_dev*y_st_dev))


def get_prob_x_greater_than_ys(x_mean, x_st_dev, y_means, y_st_devs):
    """
    :param x_mean: (float) the mean of x, which follows a normal distribution
    :param x_st_dev: (float) the standard deviation of x
    :param y_means: (list) the means of ys, which follow normal distributions
    :param y_st_devs: (list) the standard deviations of ys
    :return: the probability that x is greater than all ys (i.e., x > y_1 and x > y_2, and ...)
        (note all distributions should be independent)
    """

    prob = 1
    for i in range(len(y_means)):
        prob *= get_prob_normal_greater_0(
            mean=x_mean - y_means[i],
            st_dev=np.sqrt(x_st_dev**2 + y_st_devs[i]**2))

    return prob


if __name__ == "__main__":

    print(get_prob_normal_greater_0(-1, 1))
    print(get_prob_x_greater_than_y(x_mean=10, x_st_dev=2,
                                    y_mean=9, y_st_dev=2))
    print(get_prob_x_greater_than_ys(x_mean=10, x_st_dev=2,
                                     y_means=[9, 11], y_st_devs=[2, 2]))

    print(get_cumulative_mean([1, 2, 3]))
    print(get_cumulative_var([1, 2, 3]))


# np.seterr(all='raise')
import warnings

import matplotlib.pyplot as plt
import numpy as np
import statsmodels.api as sm
from numpy.polynomial import polynomial as P
from scipy import stats
from scipy.optimize import curve_fit
from sklearn.preprocessing import PolynomialFeatures, OneHotEncoder
from statsmodels.sandbox.regression.predstd import wls_prediction_std


# ------- Single variable regression models ----------
class _OneVarRegression:

    def __init__(self, x, y):
        self._x = x
        self._y = y
        self._coeffs = None

    def get_coeffs(self):
        return self._coeffs

    def get_predicted_y(self, x):
        raise NotImplementedError

    def get_derivative(self, x):
        raise NotImplementedError

    def get_roots(self):
        raise NotImplementedError

    def plot_fit(self, x_range=None, fig_size=None):

        fig, ax = plt.subplots(figsize=fig_size)
        self.add_to_ax(ax=ax, x_range=x_range)
        fig.show()

    def add_to_ax(self, ax, x_range=None):

        if x_range:
            xs = np.linspace(x_range[0], x_range[-1], 50)
        else:
            xs = np.linspace(self._x[0], self._x[-1], 50)

        ys = self.get_predicted_y(xs)
        ax.scatter(self._x, self._y)
        ax.plot(xs, ys, color='red')


class PolyRegression(_OneVarRegression):
    # regression of form: f(x) = c0 + c1*x + c2*x^2 + c3*x^3 + ... + cn*x^n

    def __init__(self, x, y, degree=1):

        _OneVarRegression.__init__(self, x, y)
        self._coeffs = P.polyfit(x=x, y=y, deg=degree)

    def get_coeffs(self):
        return self._coeffs

    def get_predicted_y(self, x):

        f = 0
        for i, coeff in enumerate(self._coeffs):
            f += coeff * pow(x, i)
        return f

    def get_derivative(self, x):

        result = 0
        for i in range(len(self._coeffs) - 1):
            result += (i + 1) * self._coeffs[i + 1] * pow(x, i)

        return result

    def get_roots(self):

        return P.polyroots(self._coeffs)


class ExpRegression (_OneVarRegression):
    # regression of form f(x) = c0 + c1*exp(c2*x)

    def __init__(self, x, y, if_c0_zero=False, p0=None):

        _OneVarRegression.__init__(self, x, y)
        self._ifC0Zero = if_c0_zero
        if if_c0_zero:
            self._coeffs, cov = curve_fit(self.exp_func_c0_zero, x, y, p0=p0) #jac=self.Jac_c0_zero)
        else:
            self._coeffs, cov = curve_fit(self.exp_func, x, y, p0=p0) # jac=self.Jac)

    def get_predicted_y(self, x):
        if self._ifC0Zero:
            return self.exp_func_c0_zero(x, *self._coeffs)
        else:
            return self.exp_func(x, *self._coeffs)

    def get_derivative(self, x):
        if self._ifC0Zero:
            return self.der_exp_func_c0_zero(x, *self._coeffs)
        else:
            return self.der_exp_func(x, *self._coeffs)

    @staticmethod
    def exp_func(x, c0, c1, c2):
        return c0 + c1 * np.exp(c2 * x)

    @staticmethod
    def exp_func_c0_zero(x, c1, c2):
        return c1 * np.exp(c2 * x)

    @staticmethod
    def der_exp_func(x, c0, c1, c2):
        return c1 * c2 * np.exp(c2 * x)

    @staticmethod
    def der_exp_func_c0_zero(x, c1, c2):
        return c1 * c2 * np.exp(c2 * x)

    @staticmethod
    def Jac(x, c0, c1, c2):
        v = np.exp(c1*x)
        return np.array([np.ones(len(x)), v, c2*x*v]).transpose()

    @staticmethod
    def Jac_c0_zero(x, c1, c2):
        v = np.exp(c1*x)
        return np.array([v, c2*x*v]).transpose()


class PowerRegression (_OneVarRegression):
    # regression of form f(x) = c0 + c1*pow(x, c3)

    def __init__(self, x, y, if_c0_zero=False, p0=None):
        _OneVarRegression.__init__(self, x, y, )
        self._ifC0Zero = if_c0_zero
        if if_c0_zero:
            self._coeffs, cov = curve_fit(self.power_func_c0_zero, x, y, p0=p0)
        else:
            self._coeffs, cov = curve_fit(self.power_func, x, y, p0=p0)

    def get_coeffs(self):
        return self._coeffs

    def get_predicted_y(self, x):
        if self._ifC0Zero:
            return self.power_func_c0_zero(x, *self._coeffs)
        else:
            return self.power_func(x, *self._coeffs)

    @staticmethod
    def power_func(x, c0, c1, c2):
        return c0 + c1 * np.power(x, c2)

    @staticmethod
    def power_func_c0_zero(x, c1, c2):
        return c1 * np.power(x, c2)


# ------- Single variable polynomial model with inference ----------
class SingleVarPolyRegWithInference:

    # for additional information:
    # http://markthegraph.blogspot.com/2015/05/using-python-statsmodels-for-ols-linear.html

    def __init__(self, x, y, degree=1):

        self.degree = degree
        self.x = x
        self.X = self.get_X(x)
        self.y = y
        self.fitted = sm.OLS(self.y, self.X).fit()

    def get_predicted_y(self, x_pred):
        """ :returns predicted y values at the provided x values """

        X_pred = self.get_X(x_pred)
        return self.fitted.predict(X_pred)

    def get_X(self, x):
        """
        :param x: (list or np.array) observed x values (x1, x2, x3, ..., xn)
        :return: (matrix) of X = [(1, x1'),
                                  (1, x2'),
                                  (1, x3'),
                                  ...,
                                  (1, xn')]
        """

        if self.degree == 1:
            # for degree 1, we only need to add constant vector 1 to x to get
            # [1, x1]
            # [1, x2] ...
            return sm.add_constant(x)
        else:
            col_x = [x]       # to build [x, x^2, x^3, ...]
            for i in range(2, self.degree+1):
                col_x.append(np.power(x, i))
            # turn [x, x^2, x^3, ...] to X
            X = np.column_stack(col_x)
            return sm.add_constant(X)

    def get_predicted_y_CI(self, x_pred, alpha=0.05):
        """ :returns confidence interval of the predicted y at the provided x values """

        # http://www2.stat.duke.edu/~tjl13/s101/slides/unit6lec3H.pdf
        y_hat = self.fitted.predict(self.X)     # predicted y at X
        y_err = self.y - y_hat                  # residuals
        mean_x = self.x.mean()                  # mean of x
        n = len(self.X)                         # number of observations
        dof = n - self.fitted.df_model - 1      # degrees of freedom
        t = stats.t.ppf(1 - alpha/2, df=dof)      # t-statistics
        s_err = np.sum(np.power(y_err, 2))      # sum of squared error
        conf = t * np.sqrt((s_err / (n - 2)) * (1.0 / n + (np.power((x_pred - mean_x), 2) /
                                                           ((np.sum(np.power(x_pred, 2))) - n * (
                                                               np.power(mean_x, 2))))))
        y_pred = self.get_predicted_y(x_pred)
        upper = y_pred + abs(conf)
        lower = y_pred - abs(conf)

        return lower, upper

    def get_predicted_y_PI(self, x_pred, alpha=0.05):
        """ :returns prediction interval of the y at the provided x values """
        X_pred = self.get_X(x_pred)
        sdev, lower, upper = wls_prediction_std(self.fitted, exog=X_pred, alpha=alpha)
        return lower, upper

    def get_coeffs(self):
        """ :returns coefficients of the fitted model """

        return self.fitted.params

    def get_pvalues(self):
        """ :returns p-values of coefficients """
        return self.fitted.pvalues

    def get_derivative(self, x):
        """ :returns derivative of the polynomial function at x """

        coeffs = self.fitted.params
        result = 0
        for i in range(len(coeffs)-1):
            result += (i+1) * coeffs[i+1] * pow(x, i)

        return result

    def get_zero(self):
        """
        :return: x for which f(x) = 0
        """
        coeffs = np.fliplr([self.fitted.params])[0]
        if np.linalg.norm(coeffs) == 0:
            return np.zeros(len(coeffs))
        else:
            return np.roots(coeffs)


# ------- Multi-variable linear regression ----------
class LinearRegression:
    
    def __init__(self, l2_penalty=0):
        """
        :param l2_penalty: (float) l2 regularization penalty
        """

        self.l2Penalty = l2_penalty
        self._coeffs = None

    def fit(self, X, y, forgetting_factor=1):

        # W
        N = len(y)
        if forgetting_factor < 1:
            w = []
            for i in range(N):
                w.append(pow(forgetting_factor, N - i - 1))
            W = np.diag(w)
        else:
            W = np.diag([1]*N)

        # XT.W.X
        XTX = np.transpose(X) @ W @ X
        # XT.W.y
        XTy = np.transpose(X) @ W @ y

        if self.l2Penalty > 0:
            self._add_l2(XTX)

        # solve to estiamte the coefficients
        self._coeffs = np.linalg.solve(XTX, XTy)

    def get_coeffs(self):
        return self._coeffs

    def set_coeffs(self, values):
        self._coeffs = values

    def get_y(self, x):
        return x @ self._coeffs

    def _add_l2(self, XTX):
        I = np.identity(XTX.shape[0])
        XTX += I * self.l2Penalty


class RecursiveLinearReg(LinearRegression):

    def __init__(self, l2_penalty=0):
        """
        :param l2_penalty: (float) l2 regularization penalty
        """

        LinearRegression.__init__(self, l2_penalty=l2_penalty)

        self.itr = 0
        self._X = None
        self._y = None
        self._B = None
        self._H = None
        self._w = None

    def update(self, x, y, forgetting_factor=1):

        self.itr += 1

        if self.itr < 2*len(x):
            if self._X is None:
                self._X = np.array(x)
            else:
                self._X = np.vstack((self._X, x))
            if self._y is None:
                self._y = np.array(y)
            else:
                self._y = np.vstack((self._y, y))
            if self._w is None:
                self._w = np.array(1.0)
            else:
                self._w *= forgetting_factor
                self._w = np.append(self._w, 1.0)

        elif self.itr == 2*len(x):
            self._X = np.vstack((self._X, x))
            self._y = np.vstack((self._y, y))
            self._w *= forgetting_factor
            self._w = np.append(self._w, 1.0)
            W = np.diag(self._w)

            # XTX
            XTX = np.transpose(self._X) @ W @ self._X

            # add L2 regularization
            if self.l2Penalty > 0:
                self._add_l2(XTX)

            # B = (XT.X)-1
            try:
                self._B = np.linalg.inv(XTX)
            except np.linalg.LinAlgError:
                raise ValueError('Inverse cannot be calculated. Add an l2 regularization penalty.')

            # theta = B.XT.y
            self._coeffs = np.transpose(self._B @ np.transpose(self._X) @ W @ self._y)[0]
        else:
            # turn x into a column vector
            x = np.atleast_2d(np.array(x)).T
            # gamma = lambda + xT.B.x
            gamma = (forgetting_factor + np.transpose(x) @ self._B @ x)[0]
            if gamma <= forgetting_factor:
                warnings.warn('Unstable regression, iteration: {}.'.format(self.itr))

            # epsilon = y - thetaT*x
            epsilon = (y - self._coeffs @ x)[0]
            # theta = theta + B.x.epsilon/gamma
            self._coeffs += (self._B @ x * epsilon / gamma).flatten()
            # B = (B-B.x.xT.B/gamma)/lambda
            d = self._B @ x @ np.transpose(x) @ self._B
            self._B -= d / gamma

            self._B /= forgetting_factor


# ------- Q-functions ----------
class _QFunction:

    def __init__(self, name=None):
        self.name = name

    def update(self, f, x_cont, x_bin=None, forgetting_factor=1):
        raise NotImplementedError

    def f(self, x_cont, x_bin=None):
        raise NotImplementedError


class PolynomialQFunction(_QFunction):
    """ polynomial Q-function """

    def __init__(self, name=None, degree=2, l2_penalty=0):
        """
        :param name: (string) name of this q-function
        :param degree: (int) degree of the polynomial function
        :param l2_penalty: (float) l2 regularization penalty
        """

        _QFunction.__init__(self, name=name)

        # to find row data to fit the polynomial function
        self.poly = PolynomialFeatures(degree=degree)
        self.oneHotEncoder = OneHotEncoder()

        # recursive linear regression
        self.reg = RecursiveLinearReg(l2_penalty=l2_penalty)

    def _get_x(self, x_cont, x_bin=None):
        """
        use the values of the continuous and categorical features to find the row data to use for fitting
        a linear regression
        :param x_cont: (list) of values for continuous features
        :param x_bin: (list) of values for indicator features (only 0 and 1 values)
        :return: the row data to use for fitting a linear regression
        """

        if x_cont is None or len(x_cont) == 0:
            # if the values of continuous features are not provided, we assume 1 continuous feature
            # with value set to 0. This is to make sure we can calculate the intercept.
            x_cont = [0]

        x_continuous = []
        if x_cont is not None:
            x_cont = np.atleast_1d(x_cont)
            x_continuous = self.poly.fit_transform(X=[x_cont])[0]

        if x_bin is not None and len(x_bin) > 0:
            x_indicator = np.atleast_1d(x_bin)

            # update the regressors based on the value of indicator features
            for i in x_indicator:
                x_continuous = np.append(x_continuous, i * x_continuous)

        return x_continuous

    def update(self, y, x_cont=None, x_bin=None, forgetting_factor=1):
        """
        updates the fitted Q-function
        :param y: the observed value of the Q-function at the given feature values
        :param x_cont: (list) of values for continuous features
        :param x_bin: (list) of values for indicator features (can take only 0 or 1)
        :param forgetting_factor: (float) forgetting factor
        """

        self.reg.update(x=self._get_x(x_cont=x_cont, x_bin=x_bin),
                        y=y,
                        forgetting_factor=forgetting_factor)

    def f(self, x_cont, x_bin=None):
        """
        :param x_cont: (list) of values for continuous features
        :param x_bin: (list) of values for indicator features (only takes 0 or 1)
        :return: the value of Q-function at the provided features
        """
        return self.reg.get_y(x=self._get_x(
            x_cont=x_cont,
            x_bin=x_bin))

    def get_coeffs(self):
        """
        :return: the coefficients of the fitted regression model
        """
        return self.reg.get_coeffs()

    def set_coeffs(self, values):
        """ set the values of the coefficients of the fitted regression model
        :param values: (list) coefficient values
        """
        self.reg.set_coeffs(values=values)



import enum

import numpy as np
from numba import jit
from scipy.linalg import logm

from deampy.random_variates import Empirical, Exponential, Multinomial


@jit(nopython=True)  # nopython=True makes it faster (forces full compilation)
def _out_rate(rates, idx):
    """
    :param rates: list of rates leaving this state
    :param inx: index of this state
    :returns the rate of leaving this state (the sum of rates)
    """

    sum_rates = 0
    for i, v in enumerate(rates):
        if i != idx:
            sum_rates += v
    return sum_rates


@jit(nopython=True)
def condense_prob_matrix(transition_prob_matrix):
    """
    Creates a condensed transition probability matrix where transitions with zero probabilities are removed.
    :param transition_prob_matrix: (list of lists) transition probability matrix
    :return: (tuple) (matrix of nonzero transition probability, matrix of indices with nonzero transition probabilities)
    """

    non_zero_probs = []
    indices_non_zero_probs = []
    for i, probs in enumerate(transition_prob_matrix):
        # find the indices of non-zero probabilities
        row_non_zero_probs = []
        row_non_zero_probs_indices = []
        for j, p in enumerate(probs):
            if p > 0:
                row_non_zero_probs.append(p)
                row_non_zero_probs_indices.append(j)

        non_zero_probs.append(row_non_zero_probs)
        indices_non_zero_probs.append(row_non_zero_probs_indices)

    return non_zero_probs, indices_non_zero_probs


def _assert_prob_matrix(prob_matrix):
    """
    :param prob_matrix: (list of lists or np.ndarray) transition probability matrix
    :return: None if the matrix is valid, otherwise raise an error
    """

    if len(prob_matrix) == 0:
        raise ValueError('An empty probability matrix is provided.')

    assert isinstance(prob_matrix, (list, np.ndarray)), \
        "prob_matrix is a matrix that should be represented as a list of lists or numpy array."

    for i, row in enumerate(prob_matrix):
        assert isinstance(row, (list, np.ndarray)), \
            'prob_matrix should be a list of lists or numpy array. ' \
            'Row {} is not a list or numpy array.'.format(i)

    for i, probs in enumerate(prob_matrix):

        # check if the sum of probabilities in this row is 1.
        s = sum(probs)
        if s < 0.99999 or s > 1.00001:
            raise ValueError('Sum of each row in a probability matrix should be 1. '
                             'Sum of row {} is {}.'.format(i, s))

    if isinstance(prob_matrix, list):
        prob_matrix = np.array(prob_matrix)

    return prob_matrix


def _assert_rate_matrix(rate_matrix):
    """
    :param rate_matrix: (list of lists or np.ndarray) transition rate matrix
    :return: None if the matrix is valid, otherwise raise an error
    """

    if len(rate_matrix) == 0:
        raise ValueError('An empty rate matrix is provided.')

    assert isinstance(rate_matrix, (list, np.ndarray)), \
        "rate_matrix is a matrix that should be represented as a list of lists or numpy array."

    for i, row in enumerate(rate_matrix):
        assert isinstance(row, (list, np.ndarray)), \
            'rate_matrix should be a list of lists or numpy array. ' \
            'Row {} is not a list or numpy array.'.format(i)

    for i, row in enumerate(rate_matrix):
        # make sure all non-diagonal rates are non-negative
        for j, value in enumerate(row):

            if i != j and (value is None or value <0):
                raise ValueError('All non-diagonal rates in a transition rate matrix should be non-negative. '
                                 'Negative rate ({}) found in row index {}.'.format(value, i))
            elif i == j:
                # make sure diagonal rates are None or 0
                if value is None:
                    row[j] = 0.0
                if value < 0:
                    if abs(sum(row)) > 0.00001:
                        raise ValueError(
                            'The diagonal rate in row {} is negative; therefore the sum of rates in that row should be zero.'.format(i))

    if isinstance(rate_matrix, list):
        rate_matrix = np.array(rate_matrix)

    return rate_matrix


def _assert_dynamic_matrix(rate_matrix):

    # check if rate_matrix has function get_matrix()
    if not hasattr(rate_matrix, 'get_matrix'):
        raise ValueError("The rate_matrix should have a function 'get_matrix' to get the matrix values.")


def get_prob_2_transitions(rates_out, trans_rate_matrix, delta_t):

    # probability that transition occurs within delta_t for each state
    probs_out = []
    for rate in rates_out:
        probs_out.append(1 - np.exp(-delta_t * rate))

    # calculate the probability of two transitions within delta_t for each state
    prob_out_out = []
    for i in range(len(trans_rate_matrix)):

        # probability of leaving state i withing delta_t
        prob_out_i = probs_out[i]

        # probability of leaving the new state after i withing delta_t
        prob_out_again = 0

        for j in range(len(trans_rate_matrix[i])):
            if not i == j:

                # probability of transition from i to j
                prob_i_j = 0
                if rates_out[i] > 0:
                    prob_i_j = trans_rate_matrix[i][j] / rates_out[i]
                # probability of transition from i to j and then out of j within delta_t
                prob_i_j_out = prob_i_j * probs_out[j]
                # update the probability of transition out of i and again leaving the new state
                prob_out_again += prob_i_j_out

        # store the probability of leaving state i to a new state and leaving the new state withing delta_t
        prob_out_out.append(prob_out_i * prob_out_again)

    # return the maximum probability of two transitions within delta_t
    return max(prob_out_out)


@jit(nopython=True)
def _continuous_to_discrete_main(trans_rate_matrix, delta_t):

    # list of rates out of each row
    rates_out = []
    for i, row in enumerate(trans_rate_matrix):
        rates_out.append(_out_rate(row, i))

    prob_matrix = np.zeros((len(trans_rate_matrix), len(trans_rate_matrix[0])))
    # prob_matrix = []
    for i in range(len(trans_rate_matrix)):
        # calculate probabilities
        # prob_matrix.append([0] * len(trans_rate_matrix[i]))
        for j in range(len(trans_rate_matrix[i])):
            if i == j:
                prob_matrix[i][j] = float(np.exp(-rates_out[i] * delta_t))
            else:
                if rates_out[i] > 0:
                    prob_matrix[i][j] = float((1 - np.exp(-rates_out[i] * delta_t)) * trans_rate_matrix[i][j] / rates_out[i])

    # # calculate the maximum probability of two transitions within delta_t
    # max_prob_2_transitions = get_prob_2_transitions(
    #     rates_out=rates_out, trans_rate_matrix=trans_rate_matrix, delta_t=delta_t)

    # return the probability matrix and the upper bound for the probability of two transitions with delta_t
    return prob_matrix


def continuous_to_discrete(trans_rate_matrix, delta_t):
    """
    :param trans_rate_matrix: (list of lists) transition rate matrix (assumes None or 0 for diagonal elements)
    :param delta_t: cycle length
    :return: transition probability matrix (list of lists)
             and the upper bound for the probability of two transitions within delta_t (float)
        converting [p_ij] to [lambda_ij] where
            mu_i = sum of rates out of state i
            p_ij = exp(-mu_i*delta_t),      if i = j,
            p_ij = (1-exp(-mu_i*delta_t))*lambda_ij/mu_i,      if i != j.
    """

    # error checking
    trans_rate_matrix = _assert_rate_matrix(trans_rate_matrix)

    return _continuous_to_discrete_main(trans_rate_matrix=trans_rate_matrix, delta_t=delta_t)


@jit(nopython=True)
def _discrete_to_continuous_main(trans_prob_matrix, delta_t):

    rate_matrix = []
    for i, row in enumerate(trans_prob_matrix):
        rate_row = []  # list of rates
        # calculate rates
        for j in range(len(row)):
            # rate is None for diagonal elements
            if i == j:
                rate = None
            else:
                # rate is zero if this is an absorbing state
                if trans_prob_matrix[i][i] == 1:
                    rate = 0
                else:
                    rate = float(-np.log(trans_prob_matrix[i][i]) * trans_prob_matrix[i][j] / (
                                (1 - trans_prob_matrix[i][i]) * delta_t))
            # append this rate
            rate_row.append(rate)
        # append this row of rates
        rate_matrix.append(rate_row)

    return rate_matrix


def discrete_to_continuous(trans_prob_matrix, delta_t, method='approx'):
    """
    :param trans_prob_matrix: (list of lists) transition probability matrix
    :param delta_t: cycle length
    :param method: method to convert transition probability matrix to transition rate matrix
                   'log': use the matrix logarithm method (default)
                   'approx': use the approximation method
    :return: (list of lists) transition rate matrix
        The approximation method uses:
        Converting [p_ij] to [lambda_ij] where
            lambda_ii = None, and
            lambda_ij = -ln(p_ii) * p_ij / ((1-p_ii)*Delta_t)
    """

    # error checking
    trans_prob_matrix = _assert_prob_matrix(trans_prob_matrix)
    if method == 'approx':
        return _discrete_to_continuous_main(trans_prob_matrix, delta_t)

    elif method == 'log':
        # use the matrix logarithm method

        rate_matrix = logm(trans_prob_matrix)/delta_t
        return rate_matrix

    else:
        raise ValueError("The method should be either 'log' or 'approx'.")


class _Markov:

    def __init__(self, matrix, state_descriptions=None):
        """
        :param state_descriptions: (Enum) description of the states in the format of Enum
        """

        self._n_states = len(matrix)

        if state_descriptions is not None:
            assert type(state_descriptions) is enum.EnumType, 'State description should be an enumeration.'
            assert len(state_descriptions) == len(matrix), \
                ('The number of states in the transition probability/rate matrix '
                 'and the state description should be equal.')

            for i, element in enumerate(state_descriptions):
                assert i == element.value, \
                    'The elements in the state description should be indexed 0, 1, 2, ...' \
                    'The state {} is indexed {} but should be indexed {}.'.format(str(element), element.value, i)

        self._ifStateDescriptionProvided = False if state_descriptions is None else True
        if self._ifStateDescriptionProvided:
            self._states = list(state_descriptions)

    def _error_check(self, current_state_index=None, current_state=None):

        if current_state_index is None and current_state is None:
            raise ValueError('Either current_state_index or current_state should be provided.')

        if current_state_index is not None:
            if not (0 <= current_state_index < self._n_states):
                raise ValueError('The value of the current state index should be greater '
                                 'than 0 and smaller than the number of states. '
                                 'Value provided for current state index is {}.'.format(current_state_index))
        elif current_state is not None:
            if not self._ifStateDescriptionProvided:
                raise ValueError('The state description is not provided; therefore, '
                                 'current_state cannot be used. Please provide current_state_index instead.')
            assert current_state in self._states, \
                'The current state is invalid and not in the state description enumeration.'
            current_state_index = current_state.value

        return current_state_index


class MarkovJumpProcess(_Markov):

    def __init__(self, transition_prob_matrix, state_descriptions=None):
        """
        :param transition_prob_matrix: (list) transition probability matrix of a discrete-time Markov model
        :param state_descriptions: (Enum) description of the states in the format of Enum
        """

        # error checking
        transition_prob_matrix = _assert_prob_matrix(transition_prob_matrix)

        _Markov.__init__(self, matrix=transition_prob_matrix, state_descriptions=state_descriptions)

        # make empirical distributions for each state
        self._empiricalDists = []
        for i, probs in enumerate(transition_prob_matrix):
            # create an empirical distribution over the future states from this state
            self._empiricalDists.append(Empirical(probabilities=probs))

        self._n_states = len(self._empiricalDists)

    def get_next_state(self, current_state_index=None, current_state=None, rng=None):
        """
        :param current_state_index: (int) index of the current state
        :param current_state: (an element of an enumeration) current state
        :param rng: random number generator object
        :return: the index of the next state or the state description of the next state
        """

        # get current state index
        current_state_index = self._error_check(current_state_index=current_state_index, current_state=current_state)

        # find the next state index by drawing a sample from
        # the empirical distribution associated with this state
        next_state_index = self._empiricalDists[current_state_index].sample(rng=rng)

        if self._ifStateDescriptionProvided:
            # if the state description is provided, return the state description
            return self._states[next_state_index]
        else:
            # return the index of the next state
            return next_state_index


class Gillespie(_Markov):

    def __init__(self, transition_rate_matrix, state_descriptions=None):
        """
        :param transition_rate_matrix: transition rate matrix of the continuous-time Markov model
        :param state_descriptions: (Enum) description of the states in the format of Enum
        """

        # error checking
        transition_rate_matrix = _assert_rate_matrix(transition_rate_matrix)

        _Markov.__init__(self, matrix=transition_rate_matrix, state_descriptions=state_descriptions)

        self._rateMatrix = transition_rate_matrix
        self._expDists = []
        self._empiricalDists = []

        for i, row in enumerate(transition_rate_matrix):

            # find sum of rates out of this state
            rate_out = _out_rate(row, i)
            # if the rate is 0, put None as the exponential and empirical distributions
            if rate_out > 0:
                # create an exponential distribution with rate equal to sum of rates out of this state
                self._expDists.append(Exponential(scale=1/rate_out))
                # find the transition rates to other states
                # assume that the rate into this state is 0
                rates = []
                for j, v in enumerate(row):
                    if i == j:
                        rates.append(0)
                    else:
                        rates.append(v)

                # calculate the probability of each event (prob_j = rate_j / (sum over j of rate_j)
                probs = np.array(rates) / rate_out
                # create an empirical distribution over the future states from this state
                self._empiricalDists.append(Empirical(probs))

            else:  # if the sum of rates out of this state is 0
                self._expDists.append(None)
                self._empiricalDists.append(None)

    def get_next_state(self, current_state_index=None, current_state=None, rng=None):
        """
        :param current_state_index: index of the current state.
        :param current_state: (an element of an enumeration) current state
        :param rng: random number generator object
        :return: (dt, i) where dt is the time until next event, and i is the index of the next state
                or the state description of the next state.
               It returns None for dt if the process is in an absorbing state
        """

        # get the current state index
        current_state_index = self._error_check(current_state_index=current_state_index, current_state=current_state)

        # if this is an absorbing state (i.e. sum of rates out of this state is 0)
        if self._expDists[current_state_index] is None:
            # the process stays in the current state
            dt = None
            i = current_state_index
        else:
            # find the time until next event
            dt = self._expDists[current_state_index].sample(rng=rng)
            # find the next state
            i = self._empiricalDists[current_state_index].sample(rng=rng)

        i = self._states[i] if self._ifStateDescriptionProvided else i

        return dt, i


class _CohortMarkov:

    def __init__(self):
        self._numInStates = []  # list of state sizes
        self._numInStatesOverTime = []  # list of number of patients in each state over time
        self._numToStatesOverTime = []  # list of number of transitions to each state over time
        self._currentTimeStep = 0

    def get_num_in_states(self):
        """
        :return: list of number of patients in each state
        """
        return self._numInStates


class DiscreteTimeCohortMarkov(_CohortMarkov):

    def __init__(self):

        _CohortMarkov.__init__(self)

        self._nonZeroProbs = [] # list of non-zero probabilities
        self._indicesNonZeroProbs = [] # list of state indices with non-zero probabilities

    def initialize(self, initial_condition):

        self._numInStates = initial_condition
        self._numInStatesOverTime = [[] for i in range(len(self._numInStates))]
        self._numToStatesOverTime = [[] for i in range(len(self._numInStates))]

    def condense_prob_matrix(self, transition_prob_matrix):
        """
        Condense the transition probability matrix to include only non-zero probabilities.
        """

        # condense the transition probability matrix to include only non-zero probabilities
        self._nonZeroProbs, self._indicesNonZeroProbs = condense_prob_matrix(
            transition_prob_matrix=transition_prob_matrix)

    def simulate_one_time_step(self, rng=None):

        # store the size of each state
        self.record_number_in_states()

        # initialize the number of transitions to each state
        num_to_states = [0] * len(self._numInStates)
        temp_num_in_states = self._numInStates.copy()

        for s in range(len(self._numInStates)):
            if self._numInStates[s] > 0:
                # find the number of transitions to each state
                binomial = Multinomial(N=self._numInStates[s],
                                       pvals=self._nonZeroProbs[s])
                outs = binomial.sample(rng)
                # update the number of transitions to each state
                for i in range(len(outs)):
                    if self._indicesNonZeroProbs[s][i] != s:
                        num_to_states[self._indicesNonZeroProbs[s][i]] += outs[i]
                # update the number of patients in this state
                temp_num_in_states[s] -= sum(outs)
                # update the number of patients in states
                for i in range(len(outs)):
                    temp_num_in_states[self._indicesNonZeroProbs[s][i]] += outs[i]

        # update the number of patients in each state
        self._numInStates = temp_num_in_states.copy()

        # store the number of transitions to each state
        for i in range(len(self._numInStates)):
            self._numToStatesOverTime[i].append(num_to_states[i])

        self._currentTimeStep += 1

    def record_number_in_states(self):

        # store the size of each state at the end of simulation
        for i in range(len(self._numInStates)):
            self._numInStatesOverTime[i].append(self._numInStates[i])

    def simulate(self, transition_prob_matrix, initial_condition, n_time_steps, rng=None):
        """
        :param transition_prob_matrix: (list of lists) transition probability matrix
        :param initial_condition: initial condition
        :param n_time_steps: (int) number of time steps to simulate the cohort
        :param rng: random number generator object
        """

        # error checking
        transition_prob_matrix = _assert_prob_matrix(transition_prob_matrix)

        assert len(initial_condition) == len(transition_prob_matrix), \
            'The length of the initial condition should be equal to the number of states in the transition matrix.'

        # condense the transition probability matrix to include only non-zero probabilities
        self._nonZeroProbs, self._indicesNonZeroProbs = condense_prob_matrix(
            transition_prob_matrix=transition_prob_matrix)

        # initialize the number of patients in each state
        self.initialize(initial_condition=initial_condition)

        for k in range(n_time_steps):
            self.simulate_one_time_step(rng=rng)

        # store the size of each state at the end of simulation
        self.record_number_in_states()

    def get_times(self):
        """
        :return: list of time steps where the size of each state is stored
        """
        return range(self._currentTimeStep + 1)

    def get_periods(self):
        """
        :return: list of time periods where the numbers entering each state is stored
        """
        return range(1, self._currentTimeStep+1)

    def get_state_size_over_time(self, state_index):
        """
        :return: list of state sizes
        """
        return self._numInStatesOverTime[state_index]

    def get_transition_to_states_over_time(self, state_index):
        """
        :return: list of number of transitions to each state
        """

        return self._numToStatesOverTime[state_index]

    def get_sum_size_multiple_states(self, state_indices):
        """
        :param state_indices: list of indices of states
        :return: sum of the sizes of the states in the list
        """

        sum_size = 0
        for i in state_indices:
            sum_size += self._numInStates[i]

        return sum_size

    def get_sum_size_multiple_states_over_time(self, state_indices):
        """
        :param state_indices: list of indices of states
        :return: sum of the sizes of the states in the list
        """

        sum_size = np.zeros(self._currentTimeStep + 1)
        for i in state_indices:
            sum_size += np.array(self._numInStatesOverTime[i])

        return sum_size


class ContinuousTimeCohortMarkov(_CohortMarkov):

    def __init__(self, transition_rate_matrix=None, dynamic_transition_rate_matrix=None):
        """
        :param transition_rate_matrix: (list of lists) transition rate matrix
        """

        _CohortMarkov.__init__(self)

        if transition_rate_matrix is None and dynamic_transition_rate_matrix is None:
            raise ValueError('Either transition_rate_matrix or dynamic_transition_rate_matrix should be provided.')

        if transition_rate_matrix is not None and dynamic_transition_rate_matrix is not None:
            raise ValueError('Only one of transition_rate_matrix or dynamic_transition_rate_matrix should be provided.')

        # error checking
        if transition_rate_matrix is not None:
            transition_rate_matrix = _assert_rate_matrix(transition_rate_matrix)
        if dynamic_transition_rate_matrix is not None:
            _assert_dynamic_matrix(dynamic_transition_rate_matrix)

        self.transRateMatrix = transition_rate_matrix
        self.dynamicTransRateMatrix = dynamic_transition_rate_matrix

        self.dtMarkov = DiscreteTimeCohortMarkov()  # discrete-time Markov to evaluate the continuous-time Markov process
        self.deltaT = None

    def simulate(self, initial_condition, delta_t, n_time_steps, rng=None):
        """
        :param initial_condition: (list) initial size of each state
        :param delta_t: (float) cycle length
        :param n_time_steps: (int) number of time steps to simulate the cohort
        :param rng: random number generator object
        """

        self.deltaT = delta_t

        # check if a fixed or a dynamic transition rate matrix is provided
        if self.transRateMatrix is not None:
            transition_prob_matrix = continuous_to_discrete(
                trans_rate_matrix=self.transRateMatrix,
                delta_t=delta_t)

            self.dtMarkov.simulate(
                transition_prob_matrix=transition_prob_matrix,
                initial_condition=initial_condition,
                n_time_steps=n_time_steps,
                rng=rng)

        # if a dynamic transition rate matrix is provided
        else:
            self.dtMarkov.initialize(initial_condition=initial_condition)
            k = 0
            while k < n_time_steps:
                # get the transition rate matrix at this time step
                rate_matrix = self.dynamicTransRateMatrix.get_matrix(
                    num_in_states=self.dtMarkov.get_num_in_states(), time=k*delta_t)

                transition_prob_matrix = continuous_to_discrete(
                    trans_rate_matrix=rate_matrix,
                    delta_t=delta_t)

                self.dtMarkov.condense_prob_matrix(transition_prob_matrix=transition_prob_matrix)

                # simulate one time step
                self.dtMarkov.simulate_one_time_step(rng=rng)

                k += 1

            # store the size of each state at the end of simulation
            self.dtMarkov.record_number_in_states()

    def get_times(self):
        """
        :return: list of time points where the size of each state is stored
        """
        return [i*self.deltaT for i in self.dtMarkov.get_times()]

    def get_periods(self):
        """
        :return: list of time periods where the numbers entering each state is stored
        """
        return self.dtMarkov.get_periods()

    def get_state_size_over_time(self, state_index):
        """
        :return: list of state sizes
        """
        return self.dtMarkov.get_state_size_over_time(state_index)

    def get_transition_to_states_over_time(self, state_index):
        """
        :return: list of number of transitions to each state
        """
        return self.dtMarkov.get_transition_to_states_over_time(state_index)

    def get_sum_size_multiple_states(self, state_indices):
        """
        :param state_indices: list of state indices
        :return: sum of people in the specified states
        """
        return self.dtMarkov.get_sum_size_multiple_states(state_indices)

    def get_sum_size_multiple_states_over_time(self, state_indices):
        """
        :param state_indices: list of state indices
        :return: sum of people in the specified states
        """
        return self.dtMarkov.get_sum_size_multiple_states_over_time(state_indices)


import multiprocessing as mp
from math import pow

import matplotlib.pyplot as plt
import numpy as np

from deampy.format_functions import format_number
from deampy.in_out_functions import write_csv, read_csv_rows, write_dictionary_to_csv
from deampy.optimization.opt_support import *
from deampy.plots.plot_support import output_figure
from deampy.regression_models import PolynomialQFunction
from deampy.support.misc_functions import get_moving_average

MAX_PROCESSES = mp.cpu_count()  # maximum number of processors


class SimModel:
    # abstract class to be overridden by the simulation model to optimize
    # a simulation model should have the following class attributes

    def __init__(self):
        pass

    def set_approx_decision_maker(self, approx_decision_maker):
        """ to allow the optimization algorithm to set a decision maker for the model that makes
            approximately optimal decisions. """
        raise NotImplementedError

    def simulate(self, itr):
        """ to allow th optimization algorithm to get one replication of the simulation model
        :param itr: (int) the iteration of the optimization algorithm
        """
        raise NotImplementedError

    def get_seq_of_costs(self):
        """ to allow tht optimization algorithm to get the sequence of cost observed
            during the decision periods of the simulation """
        raise NotImplementedError


class State:
    # a state of the approximate policy iteration algorithm
    # contains (feature values, action combination over the next period, cost over the next period)

    def __init__(self, action_combo, cost, values_of_continuous_features, values_of_indicator_features=None):
        """
        :param values_of_continuous_features: (list) of values of continuous features
        :param values_of_indicator_features: (list) of values of indicator features
        :param action_combo: (list) of on/off status of actions for the next period
        :param cost: (float) cost of this period
        """

        self.valuesOfContinuousFeatures = values_of_continuous_features
        self.valuesOfIndicatorFeatures = values_of_indicator_features
        self.actionCombo = action_combo
        self.cost = cost
        self.costToGo = 0


class _ApproxDecisionMaker:
    # super class for epsilon-greedy and greedy decisions maker

    def __init__(self, num_of_actions):
        """
        :param num_of_actions: (int) number of actions with on/off switches
        """

        self.nOfActions = num_of_actions
        self.nOfActionCombos = int(pow(2, self.nOfActions))
        self.qFunctions = []

        self.seq_of_continuous_feature_values = [] # sequence of continuous feature values throughout the simulation
        self.seq_of_indicator_feature_values = []  # sequence of indicator feature values throughout the simulation
        self.seq_of_action_combos = []  # sequence of action combinations throughout the simulation

    def reset_for_new_iteration(self):
        """ clear the sequence of feature values and action combinations """

        self.seq_of_continuous_feature_values.clear()
        self.seq_of_indicator_feature_values.clear()
        self.seq_of_action_combos.clear()

    def make_a_decision(self, continuous_feature_values=None, indicator_feature_values=None):
        raise NotImplementedError

    def _make_a_greedy_decision(self, continuous_feature_values=None, indicator_feature_values=None):
        """ makes a greedy decision given the feature values
        :param continuous_feature_values: (list) of values for continuous features
        :param indicator_feature_values: (list) of values for indicator features (can take only 0 or 1)
        :returns (list of 0s and 1s) the selected combinations of actions
        """

        if continuous_feature_values is None and indicator_feature_values is None:
            raise ValueError('Values of either continuous or indicator features should be provided.')

        minimum = float('inf')
        opt_action_combo = None
        for i in range(self.nOfActionCombos):

            # if the q-function is not updated with any data yet
            if self.qFunctions[i].get_coeffs() is None:
                q_value = 0
            else:
                q_value = self.qFunctions[i].f(x_cont=continuous_feature_values,
                                               x_bin=indicator_feature_values)

            if q_value < minimum:
                minimum = q_value
                opt_action_combo = action_combo_of_an_index(i)

        return opt_action_combo


class GreedyApproxDecisionMaker(_ApproxDecisionMaker):
    # class to make greedy decisions

    def __init__(self, num_of_actions, q_function_degree, q_functions_csv_file):
        """
        :param num_of_actions: (int) number of actions with on/off switches
        :param q_function_degree: (float) degree of the polynomial function used for q-functions
        :param q_functions_csv_file: (string) csv filename to read the coefficient of the q-functions
        """
        _ApproxDecisionMaker.__init__(self, num_of_actions=num_of_actions)

        # read the q-functions
        rows = read_csv_rows(file_name=q_functions_csv_file, if_ignore_first_row=False, if_convert_float=True)
        # create the q-functions
        for i, row in enumerate(rows):
            q_function = PolynomialQFunction(name='Q-function for ' + str(action_combo_of_an_index(i)),
                                             degree=q_function_degree)
            # read coefficients
            # [1:-1] is to remove left and right brackets from the list of coefficients
            q_function.set_coeffs(np.fromstring(row[1][1:-1], sep=' '))
            self.qFunctions.append(q_function)

    def make_a_decision(self, continuous_feature_values=None, indicator_feature_values=None):
        """ make a greedy decision
        :returns (list of 0s and 1s) the greedy selection of actions
        """

        self.seq_of_continuous_feature_values.append(continuous_feature_values)
        self.seq_of_indicator_feature_values.append(indicator_feature_values)

        a = self._make_a_greedy_decision(continuous_feature_values=continuous_feature_values,
                                         indicator_feature_values=indicator_feature_values)

        self.seq_of_action_combos.append(a)

        return a


class EpsilonGreedyApproxDecisionMaker(_ApproxDecisionMaker):
    # class to make epsilon-greedy decisions

    def __init__(self, num_of_actions, exploration_rule, q_function_degree, l2_penalty):
        """
        :param num_of_actions: (int) number of actions with on/off switches
        :param exploration_rule: exploration rule
        :param q_function_degree: (float) degree of the polynomial function used for q-functions
        :param l2_penalty: (float) l2 regularization penalty
        """

        _ApproxDecisionMaker.__init__(self, num_of_actions=num_of_actions)

        self.explorationRule = exploration_rule
        self.rng = np.random.RandomState(seed=0)
        self.itr = 0   # iteration of the algorithm (needed to calculate exploration rate)

        # create the q-functions
        for i in range(self.nOfActionCombos):
            self.qFunctions.append(
                PolynomialQFunction(name='Q-function for '+str(action_combo_of_an_index(i)),
                                    degree=q_function_degree,
                                    l2_penalty=l2_penalty)
            )

    def make_a_decision(self, continuous_feature_values=None, indicator_feature_values=None):
        """ makes an epsilon-greedy decision given the feature values
        :param continuous_feature_values: (list) of values for continuous features
        :param indicator_feature_values: (list) of values for indicator features (can take only 0 or 1)
        """

        self.seq_of_continuous_feature_values.append(continuous_feature_values)
        self.seq_of_indicator_feature_values.append(indicator_feature_values)

        if self.rng.random_sample() < self.explorationRule.get_epsilon(itr=self.itr):
            # explore
            i = self.rng.randint(low=0, high=self.nOfActionCombos)
            a = action_combo_of_an_index(i)
        else:
            # exploit
            a = self._make_a_greedy_decision(continuous_feature_values=continuous_feature_values,
                                             indicator_feature_values=indicator_feature_values)

        self.seq_of_action_combos.append(a)

        return a

    def export_q_functions(self, csv_file='q-functions.csv'):
        """ exports the coefficients of q-functions to a csv file.
         :param csv_file: (string) csv filename to store the coefficient of the q-functions
        """

        rows = []
        for q in self.qFunctions:
            rows.append([q.name, q.get_coeffs()])

        write_csv(rows=rows, file_name=csv_file)


class ApproximatePolicyIteration:
    
    def __init__(self, sim_model, num_of_actions, learning_rule, exploration_rule,
                 q_function_degree, l2_penalty, discount_factor=1, name=None):
        """
        :param sim_model (SimModel) simulation model
        :param num_of_actions: (int) number of possible actions to turn on or off
        :param learning_rule: learning rule
        :param exploration_rule: exploration rule
        :param q_function_degree: (int) degree of the polynomial function used for q-functions
        :param l2_penalty: (float) l2 regularization penalty
        :param discount_factor: (float) is 1 / (1 + interest rate)
        :param name: (string) to use in filenames and figure titles
        """

        assert hasattr(sim_model, 'set_approx_decision_maker'), \
            'sim_model should implement the attribute set_approx_decision_maker.'
        assert hasattr(sim_model, 'simulate'), \
            'sim_model should implement the attribute simulate.'
        assert hasattr(sim_model, 'get_seq_of_costs'), \
            'sim_model should implement the attribute get_seq_of_costs.'

        self.simModel = sim_model
        self.learningRule = learning_rule
        self.discountFactor = discount_factor
        self.name = name

        self.itr_i = []  # iteration indices
        self.itr_total_cost = []  # discounted total cost over iterations
        self.itr_error = []  # errors over iterations
        self.itr_forgetting_factor = []  # forgetting factor over iterations
        self.itr_exploration_rate = []  # exploration rate over iterations

        # create the approximate decision maker
        self.appoxDecisionMaker = EpsilonGreedyApproxDecisionMaker(num_of_actions=num_of_actions,
                                                                   q_function_degree=q_function_degree,
                                                                   exploration_rule=exploration_rule,
                                                                   l2_penalty=l2_penalty)

        # assign the approximate decision maker to the model
        self.simModel.set_approx_decision_maker(
            approx_decision_maker=self.appoxDecisionMaker)

    def minimize(self, n_iterations, q_function_csv_file='q-functions.csv'):
        """
        :param n_iterations: (int) number of iterations
        :param q_function_csv_file: (string) csv filename to store the coefficient of the q-functions
        """

        for itr in range(1, n_iterations + 1):

            # store iteration, exploration rate, and forgetting factor
            self.itr_i.append(itr)
            self.itr_forgetting_factor.append(self.learningRule.get_learning_rate(itr=itr))
            self.itr_exploration_rate.append(self.appoxDecisionMaker.explorationRule.get_epsilon(itr=itr))

            # update the iteration of the approximate decision maker (to calculate exploration rate)
            self.appoxDecisionMaker.itr = itr
            self.appoxDecisionMaker.reset_for_new_iteration()

            # simulate
            self.simModel.simulate(itr=itr)

            # do back-propagation
            self._back_propagate(
                itr=itr,
                seq_of_continuous_feature_values=self.appoxDecisionMaker.seq_of_continuous_feature_values,
                seq_of_indicator_feature_values=self.appoxDecisionMaker.seq_of_indicator_feature_values,
                seq_of_action_combos=self.appoxDecisionMaker.seq_of_action_combos,
                seq_of_costs=self.simModel.get_seq_of_costs())

        # export q-functions:
        self.appoxDecisionMaker.export_q_functions(csv_file=q_function_csv_file)

    def _back_propagate(self, itr,
                        seq_of_continuous_feature_values,
                        seq_of_indicator_feature_values,
                        seq_of_action_combos,
                        seq_of_costs):

        if not (len(seq_of_continuous_feature_values)
                == len(seq_of_indicator_feature_values)
                == len(seq_of_action_combos)
                == len(seq_of_costs)):
            raise ValueError('For iteration {0}, the number of past feature values, '
                             'action combinations, and costs are not equal ({1}, {2}, vs. {3}).'
                             .format(itr, len(seq_of_indicator_feature_values),
                                     len(seq_of_action_combos), len(seq_of_costs)))
        
        # if no decisions are made, back-propagation cannot be performed
        if len(seq_of_continuous_feature_values) == 0:
            self.itr_total_cost.append(None)
            self.itr_error.append(None)
            return

        # make feature/action states
        self.states = []
        for i in range(len(seq_of_continuous_feature_values)):
            self.states.append(State(values_of_continuous_features=seq_of_continuous_feature_values[i],
                                     values_of_indicator_features=seq_of_indicator_feature_values[i],
                                     action_combo=seq_of_action_combos[i],
                                     cost=seq_of_costs[i]))

        # cost of last state in this simulation run
        self.states[-1].costToGo = self.states[-1].cost

        # calculate discounted cost-to-go of states
        i = len(self.states) - 1 - 1
        while i >= 0:
            self.states[i].costToGo = self.states[i].cost \
                                      + self.discountFactor * self.states[i + 1].costToGo
            i -= 1
        
        # store the discounted total cost
        self.itr_total_cost.append(self.states[0].costToGo)

        # store error of the first period
        q_index = index_of_an_action_combo(self.states[0].actionCombo)
        if self.appoxDecisionMaker.qFunctions[q_index].get_coeffs() is None:
            self.itr_error.append(None)
        else:
            predicted_cost = self.appoxDecisionMaker.qFunctions[q_index].f(
                values_of_continuous_features=self.states[0].valuesOfContinuousFeatures,
                values_of_indicator_features=self.states[0].valuesOfIndicatorFeatures)
            self.itr_error.append(self.states[0].costToGo - predicted_cost)

        # update q-functions
        forgetting_factor = self.learningRule.get_learning_rate(itr=itr)

        for s in self.states:
            q_index = index_of_an_action_combo(s.actionCombo)
            self.appoxDecisionMaker.qFunctions[q_index].update(
                values_of_continuous_features=s.valuesOfContinuousFeatures,
                values_of_indicator_features=s.valuesOfIndicatorFeatures,
                f=s.costToGo,
                forgetting_factor=forgetting_factor)

    def export_results(self, csv_file='approximate-policy-iteration.csv'):
        """ exports the iteration of the algorithm into a csv file
        :param csv_file: csv file to export the results of iterations to
        """

        columns = dict()
        columns['iteration'] = self.itr_i
        columns['cost'] = self.itr_total_cost
        columns['error'] = self.itr_error
        columns['forgetting factor'] = self.itr_forgetting_factor
        columns['exploration rate'] = self.itr_exploration_rate

        write_dictionary_to_csv(dictionary=columns, file_name=csv_file)

    # ---------- plots -------------
    def plot_iterations(self, moving_ave_window=None, y_ranges=None, y_labels=None,
                        n_last_iterations_to_ave=None,
                        fig_size=(6, 6), filename=None):

        if y_ranges is None:
            y_ranges = [None]*4
        if y_labels is None:
            y_labels = [None]*4

        f, axarr = plt.subplots(4, 1, figsize=fig_size, sharex=True)

        # cost
        self.add_cost_itr(ax=axarr[0], moving_ave_window=moving_ave_window,
                          y_range=y_ranges[0], y_label=y_labels[0],
                          n_last_iterations_to_ave=n_last_iterations_to_ave)

        # error
        self.add_error_itr(ax=axarr[1], moving_ave_window=None,
                           y_range=y_ranges[1], y_label=y_labels[1],
                           n_last_iterations_to_ave=n_last_iterations_to_ave)

        # forgetting factor
        self.add_forgetting_factor_itr(ax=axarr[2], y_range=y_ranges[2], y_label=y_labels[2])

        # exploration rate
        self.add_exploration_rate_itr(ax=axarr[3], y_range=y_ranges[3], y_label=y_labels[3])

        f.tight_layout()
        f.align_ylabels()

        output_figure(plt=plt, file_name=filename)

    def plot_cost_itr(self, moving_ave_window=None,
                      y_range=None,
                      y_label='Discounted total cost',
                      x_label='Iteration', fig_size=(6, 5)):
        """ plots of cost as the algorithm iterates """

        fig, ax = plt.subplots(figsize=fig_size)

        self.add_cost_itr(ax=ax, moving_ave_window=moving_ave_window,
                          y_range=y_range, y_label=y_label)

        ax.set_xlabel(x_label)

        plt.show()

    def add_cost_itr(self, ax, moving_ave_window=None,
                     y_range=None, y_label=None, n_last_iterations_to_ave=None, sig_digits=4):
        """
        :return: a plot of cost as the algorithm iterates
        """

        # discounted cost
        ax.plot(self.itr_i, self.itr_total_cost)
        # moving average of the objective function
        if moving_ave_window is not None:
            ax.plot(self.itr_i, get_moving_average(self.itr_total_cost, window=moving_ave_window),
                    'k-', markersize=1)

        if n_last_iterations_to_ave is not None:
            self._add_ave_to_ax(ax=ax, data=self.itr_total_cost,
                                n_last_iterations_to_ave=n_last_iterations_to_ave,
                                sig_digits=sig_digits)

        if y_label is None:
            y_label = 'Discounted\ntotal cost'

        ax.set_ylim(y_range)
        ax.set_ylabel(y_label)

    def add_error_itr(self, ax, moving_ave_window=None,
                      y_range=None, y_label=None,
                      n_last_iterations_to_ave=None, sig_digits=4):
        """
        :return: a plot of error of first period as the algorithm iterates
        """

        # discounted cost
        ax.plot(self.itr_i, self.itr_error)
        # moving average of the objective function
        if moving_ave_window is not None:
            ax.plot(self.itr_i, get_moving_average(self.itr_error, window=moving_ave_window),
                    'k-', markersize=1)

        if n_last_iterations_to_ave is not None:
            self._add_ave_to_ax(ax=ax, data=self.itr_error,
                                n_last_iterations_to_ave=n_last_iterations_to_ave,
                                sig_digits=sig_digits)

        ax.axhline(y=0, linestyle='--', color='black', linewidth=1)

        if y_label is None:
            y_label = 'Error of\nfirst period'

        ax.set_ylim(y_range)
        ax.set_ylabel(y_label)

    def add_forgetting_factor_itr(self, ax, y_range=None, y_label=None):

        ax.plot(self.itr_i, self.itr_forgetting_factor)

        if y_label is None:
            y_label = 'Forgetting factor'
        if y_range is None:
            y_range = (0, 1)

        ax.set_ylim(y_range)
        ax.set_ylabel(y_label)

    def add_exploration_rate_itr(self, ax, y_range=None, y_label=None):

        ax.plot(self.itr_i, self.itr_exploration_rate)

        if y_label is None:
            y_label = 'Exploration'
        if y_range is None:
            y_range = (0, 1)

        ax.set_ylim(y_range)
        ax.set_ylabel(y_label)

    @staticmethod
    def _add_ave_to_ax(ax, data, n_last_iterations_to_ave, sig_digits):

        n_last_iterations_to_ave = int(n_last_iterations_to_ave)
        ave_cost = sum(data[-n_last_iterations_to_ave:]) / n_last_iterations_to_ave
        ax.text(0.99, 0.98,
                format_number(ave_cost, sig_digits=sig_digits),
                horizontalalignment='right',
                verticalalignment='top',
                transform=ax.transAxes)


class MultiApproximatePolicyIteration:
    # class to create multiple approximate policy iterations

    def __init__(self, num_of_actions, sim_models, learning_rules, exploration_rules,
                 q_function_degrees, l2_penalties, discount_factor=1):
        """
        :param num_of_actions:
        :param sim_models:
        :param learning_rules:
        :param exploration_rules:
        :param q_function_degrees:
        :param l2_penalties:
        :param discount_factor:
        """

        self.optAlgorithm = None
        self.summary = [['Learning rule', 'Exploration rule', 'Q-function degree', 'L2 penalty', 'E[Cost]']]

        i = 0
        self.optimizers = []
        for l in learning_rules:
            for e in exploration_rules:
                for d in q_function_degrees:
                    for p in l2_penalties:

                        name = '{}_{}_degree{}_penalty{}'.format(l, e, d, p)
                        self.summary.append([str(l), str(e), d, p])

                        self.optimizers.append(ApproximatePolicyIteration(
                            name=name,
                            sim_model=sim_models[i],
                            num_of_actions=num_of_actions,
                            learning_rule=l,
                            exploration_rule=e,
                            discount_factor=discount_factor,
                            q_function_degree=d,
                            l2_penalty=p))
                        i += 1

    def minimize_all(self, n_iterations, n_last_itrs_to_find_minimum=None, if_parallel=True,
                     folder_to_save_iterations=None,
                     q_functions_folder='q-functions',
                     optimal_q_functions_csvfile='optimal_q_functions.csv'):
        """
        :param n_iterations:
        :param n_last_itrs_to_find_minimum: (int)
        :param if_parallel:
        :param folder_to_save_iterations: (string) folder to store iterations from all optimization settings
        :param q_functions_folder: (string) folder to store q-functions from all optimization settings
        :param optimal_q_functions_csvfile: (string) csv file to store the coefficients of q-functions
                corresponding to the optimal optimization settings
        """

        if not if_parallel:
            for o in self.optimizers:
                o.minimize(n_iterations=n_iterations, q_function_csv_file=q_functions_folder + '/{}.csv'.format(o.name))
        else:
            # create a list of arguments for simulating the cohorts in parallel
            args = [(o, n_iterations, q_functions_folder + '/{}.csv'.format(o.name)) for o in self.optimizers]

            # simulate all cohorts in parallel
            with mp.Pool(MAX_PROCESSES) as pl:
                self.optimizers = pl.starmap(minimize_this, args)

        # export iterations
        if folder_to_save_iterations is not None:
            for o in self.optimizers:
                o.export_results(csv_filename=folder_to_save_iterations + '/{}.csv'.format(o.name))

        # find the best option
        minimum = float('inf')
        for i, o in enumerate(self.optimizers):
            ave_cost = sum(o.itr_total_cost[-n_last_itrs_to_find_minimum:])/n_last_itrs_to_find_minimum
            self.summary[i+1].extend([ave_cost])
            if ave_cost < minimum:
                minimum = ave_cost
                self.optAlgorithm = o

        self.optAlgorithm.appoxDecisionMaker.export_q_functions(csv_file=optimal_q_functions_csvfile)
        write_csv(rows=self.summary, file_name='summary.csv', directory=folder_to_save_iterations)

    def plot_iterations(self, moving_ave_window=None, y_ranges=None,
                        n_last_iterations_to_ave=None, fig_size=(5, 6),
                        folder_to_save_figures='figures'):
        """
        :param moving_ave_window: (int) number of past iterations to use to calculate moving average

        :param fig_size: (tuple)
        :param folder_to_save_figures: (string)
        """

        for o in self.optimizers:
            o.plot_iterations(moving_ave_window=moving_ave_window, y_ranges=y_ranges,
                              fig_size=fig_size, n_last_iterations_to_ave=n_last_iterations_to_ave,
                              filename=folder_to_save_figures+'/{}.png'.format(o.name))


def minimize_this(optimizer, n_iterations, csv_file):

    # simulate and return the cohort
    optimizer.minimize(n_iterations=n_iterations, q_function_csv_file=csv_file)
    return optimizer


from enum import Enum

import matplotlib.pyplot as plt


class Columns(Enum):
    """ Index of parameters in decision and chance node dictionaries. """
    COST = 0
    UTILITY = 1
    NODES = 2
    PROB = 3


class Node:
    def __init__(self, name):
        """
        :param name:(string) key of the node in node dictionaries
        """
        self.name = name        # name of this node (also the key in the node dictionaries)
        self.cumProb = 1        # probability of visiting this node
        self.cost = 0           # immediate cost of visiting this node
        self.utility = 0        # immediate utility of visiting this node
        self.eCost = 0          # expected cost of visiting this node (includes the immediate cost)
        self.eUtility = 0       # expected utility of visiting this node (includes the immediate utility)

    def get_terminal_prob(self):
        """ abstract method to be overridden in derived classes
        :returns probability of terminal nodes (if any) """
        raise NotImplementedError("This is an abstract method and needs to be implemented in derived classes.")


class DecisionNode(Node):
    def __init__(self, name, dict_decisions, dict_chances, dict_terminals):
        """
        :param name: (string) key of this decision node in the dictionary of decision nodes
        :param dict_decisions: dictionary of decision nodes
        :param dict_chances: dictionary of chance nodes
        :param dict_terminals: dictionary of terminal nodes
        """
        self.futureNodes = []  # list of future node objects

        # find if this node is in the decision dictionary
        if name in dict_decisions:
            # if found, initialize this node
            Node.__init__(self, name)
            # find the names of future nodes for this decision node
            names = dict_decisions[name][Columns.NODES.value]
            # add the future nodes
            self.futureNodes = create_future_nodes(names, dict_chances, dict_terminals)

        else:
            raise ValueError('{} is not in the decision node dictionary'.format(self.name))

    def evaluate(self):
        """ evaluates the cost and utility of this decision node """

        for node in self.futureNodes:
            node.evaluate()

    def get_cost_utility(self):
        """ returns the expected cost and health utility of each decisions
        :return: dictionary of outcomes where key = node name and value =[expected cost, expected utility]
        """
        outcomes = dict()
        for node in self.futureNodes:
            outcomes[node.name] = [node.eCost, node.eUtility]

        return outcomes

    def get_terminal_prob(self):
        """ :returns a dictionary where key = the name of terminal nodes (if any) and values = probabilities """

        terminal_prob = dict()
        for node in self.futureNodes:
            # assign the terminal probabilities to a new key (the name of this future node)
            terminal_prob[node.name] = node.get_terminal_prob()

            # error checking: sum of cumulative probabilities of terminal nodes should be 1
            s = 0
            for key in terminal_prob[node.name]:
                s += terminal_prob[node.name][key]
            assert 0.99999 < s < 1.00001, \
                'Sum of cumulative probabilities of terminal nodes should be 1. It is {} for node {}.'\
                    .format(s, node.name)

        return terminal_prob


class ChanceNode(Node):
    def __init__(self, name, dict_chances, dict_terminals):
        """
        :param name: (string) key of this chance node in the dictionary of chance nodes
        :param dict_chances: dictionary of chance nodes
        :param dict_terminals: dictionary of terminal nodes
        """
        self.futureNodes = []  # list of future node objects
        self.pFutureNodes = [] # probabilities of future nodes

        # find if this node is in the chance dictionary
        if name in dict_chances:
            # if found, initialize this node
            Node.__init__(self, name)

            self.cost = dict_chances[name][Columns.COST.value]            # find cost
            self.utility = dict_chances[name][Columns.UTILITY.value]      # find utility
            self.pFutureNodes = dict_chances[name][Columns.PROB.value]    # find probability of each future nodes

            # find the names of future nodes for this chance node
            names = dict_chances[name][Columns.NODES.value]
            # add the future nodes
            self.futureNodes = create_future_nodes\
                (names, dict_chances, dict_terminals)

        else:
            raise ValueError('{} is not in the chance node dictionary'.format(self.name))

    def evaluate(self):
        """ evaluates the cost and utility of this chance node """

        # check probabilities of future node
        s = sum(self.pFutureNodes)
        if not(0.99999 < s < 1.00001):
            raise ValueError('Sum of probabilities out of chance nodes should be 1. It is {} for node {}.' \
                .format(s, self.name))

        # calculate expected cost and utility of this node
        self.eCost = self.cost  # adding the immediate cost
        self.eUtility = self.utility  # adding the immediate utility
        i = 0  # iterator in future nodes
        for node in self.futureNodes:
            # update the cumulative probability of this future node
            node.cumProb = self.cumProb * self.pFutureNodes[i]
            # evaluate the node
            node.evaluate()
            # add the expected cost of this future node
            self.eCost += node.eCost * self.pFutureNodes[i]
            # add the expected utility of this future node
            self.eUtility += node.eUtility * self.pFutureNodes[i]
            # increment i
            i += 1

    def get_terminal_prob(self):
        """ :returns a dictionary where key = the name of terminal nodes (if any) and values = probabilities """

        terminal_prob = dict()
        for node in self.futureNodes:
            # add the terminal probabilities to this dictionary
            terminal_prob.update(node.get_terminal_prob())

        return terminal_prob


class TerminalNode(Node):
    def __init__(self, name, dict_terminals):
        """ Instantiating a terminal node
        :param name: key of this node
        :param dict_terminals: dictionary of terminal nodes
        """

        # find if this node is in the dictionary of terminal nodes
        if name in dict_terminals:
            # create the node
            Node.__init__(self, name)
            # find the cost of this node (for terminal nodes eCost = immediate cost)
            self.cost = dict_terminals[name][Columns.COST.value]
            # find the utility of this node (for terminal nodes eUtility = immediate utility)
            self.utility = dict_terminals[name][Columns.UTILITY.value]

        else:
            raise ValueError('{} is not in the terminal node dictionary'.format(self.name))

    def evaluate(self):
        """ evaluates the cost and utility of this terminal node """
        self.eCost = self.cost
        self.eUtility = self.utility

    def get_terminal_prob(self):
        """ :returns a dictionary where key = the name of this terminal node, and value = probability """

        terminal_prob = dict()
        # add a new entity (key = name of this terminal node, value = cumulative probability)
        terminal_prob[self.name] = self.cumProb

        return terminal_prob


def create_future_nodes(names, dict_chances, dict_terminals):
    """ gets the names of future nodes and return the future node objects
    :param names: names of future nodes
    :param dict_chances: dictionary of chance nodes
    :param dict_terminals: dictionary of terminal nodes
    :return: list of future nodes
    """

    future_nodes = []     # list of future nodes to return
    i = 0           # iterator in future nodes
    for name in names:
        # if this name is associated to a chance node
        if name in dict_chances:
            # create a chance node
            cn = ChanceNode(name, dict_chances, dict_terminals)
            # append this node to the collection of future nodes
            future_nodes.append(cn)

        # if this name is associated to a terminal node
        elif name in dict_terminals:
            # create a terminal node
            tn = TerminalNode(name, dict_terminals)
            # append this node to the collection of future nodes
            future_nodes.append(tn)

        i += 1

    return future_nodes


def graph_outcomes(decision_tree):
    """ plots the expected cost and expected utility of choices """

    # specify title and labels
    fig = plt.figure('cost vs. utility')
    plt.xlabel('Expected Health Utility')
    plt.ylabel('Expected Cost')

    # build the legend and add the series
    names = []  # names of choices
    for key, cost_utility in decision_tree.get_cost_utility().items():
        plt.plot(cost_utility[1], cost_utility[0], 'o')
        names.append(key)

    # plot the legend
    plt.legend(names, loc='lower right')

    # specify the range of x and y-axes
    plt.xlim([0,1])
    plt.ylim(bottom=0)

    # show the figure
    plt.show()



from math import log, cos, pi

import numpy as np
from numpy import exp, pi, cos

from deampy.random_variates import Beta as B
from deampy.random_variates import Dirichlet as Dirich
from deampy.random_variates import Gamma as G
from deampy.random_variates import Multinomial as Mult
from deampy.random_variates import Uniform as U
from deampy.random_variates import UniformDiscrete as UD


class _Parameter:
    # super class for parameters
    def __init__(self, id=None, name=None, if_time_dep=False):
        """
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        :param if_time_dep: (bool) if the value of this parameter changes by time
        """
        self.id = id
        self.name = name
        self.value = None
        self.ifTimeDep = if_time_dep

    def sample(self, rng=None, time=None):
        """
        :param rng: optional random number generator
        :param time: optimal time
        :return: a sample from this paramter
        """
        pass

    @staticmethod
    def if_any_time_dep(parameters):

        # if this is a time-dependent parameter
        for p in parameters:
            if p.ifTimeDep:
                return True

        return False


class _SingleVariate(_Parameter):
    pass


class _MultiVariate(_Parameter):
    pass


class Constant(_SingleVariate):
    def __init__(self, value, id=None, name=None):
        """
        :param value: (float) constant value of this parameter
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """

        _SingleVariate.__init__(self, id=id, name=name)
        self.value = value

    def sample(self, rng=None, time=None):
        return self.value


class ConstantArray(_MultiVariate):
    def __init__(self, values, id=None, name=None):
        """
        :param values: (list) of constant values of this parameter
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """

        _MultiVariate.__init__(self, id=id, name=name)
        self.value = values

    def sample(self, rng=None, time=None):
        return self.value


class Uniform(_SingleVariate):
    def __init__(self, minimum=0, maximum=1, id=None, name=None):
        """
        :param minimum: (float) minimum value of a parameter with uniform distribution
        :param maximum: (float) maximum value of a parameter with uniform distribution
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name)
        self.par = U(scale=maximum-minimum, loc=minimum)

    def sample(self, rng=None, time=None):
        self.value = self.par.sample(rng=rng)
        return self.value


class UniformDiscrete(_SingleVariate):
    def __init__(self, minimum=0, maximum=1, id=None, name=None):
        """
        :param minimum: (float) inclusive minimum value
        :param maximum: (float) inclusive maximum value
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name)
        self.par = UD(l=minimum, u=maximum)

    def sample(self, rng=None, time=None):
        self.value = self.par.sample(rng=rng)
        return self.value


class Beta(_SingleVariate):
    def __init__(self, mean, st_dev, minimum=0, maximum=1, id=None, name=None):
        """
        :param mean: (float) mean value of a parameter with beta distribution
        :param st_dev: (float) st_dev value of a parameter with beta distribution
        :param minimum: (float) minimum value of a parameter with beta distribution
        :param maximum: (float) maximum value of a parameter with beta distribution
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """

        if not (minimum < mean < maximum):
            raise ValueError('Mean should be between minimum and maximum.')

        _Parameter.__init__(self, id=id, name=name)
        fit_results = B.fit_mm(mean=mean, st_dev=st_dev, minimum=minimum, maximum=maximum)
        self.par = B(a=fit_results['a'], b=fit_results['b'], loc=fit_results['loc'], scale=fit_results['scale'])

    def sample(self, rng=None, time=None):
        self.value = self.par.sample(rng=rng)
        return self.value


class Gamma(_SingleVariate):
    def __init__(self, mean, st_dev, id=None, name=None):
        """
        :param mean: (float) mean value of a parameter with beta distribution
        :param st_dev: (float) st_dev value of a parameter with beta distribution
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """

        _Parameter.__init__(self, id=id, name=name)
        fit_results = G.fit_mm(mean=mean, st_dev=st_dev)
        self.par = G(shape=fit_results['shape'], scale=fit_results['scale'])

    def sample(self, rng=None, time=None):
        self.value = self.par.sample(rng=rng)
        return self.value


class Equal(_SingleVariate):
    # value = value of another parameter

    def __init__(self, par, id=None, name=None):
        """
        :param par: (Parameter) another parameter to set this parameter equal to
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """

        _Parameter.__init__(self, id=id, name=name,
                            if_time_dep=par.ifTimeDep)

        self.par = par

    def sample(self, rng=None, time=None):

        self.value = self.par.value
        return self.value


class Multinomial(_MultiVariate):
    def __init__(self, par_n, p_values, id=None, name=None):
        """
        :param par_n: (Parameter) number of trials
        :param p_values: (list) probabilities of success for each category
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """

        if not (0.9999999 <= sum(p_values) <= 1.0000001):
            raise ValueError("Sum of p_values should be 1 (not {}) for parameter '{}'.".format(sum(p_values), name))

        _Parameter.__init__(self, id=id, name=name)
        self.parN = par_n
        self.pVals = p_values

    def sample(self, rng=None, time=None):

        self.value = Mult(N=self.parN.value, pvals=self.pVals).sample(rng=rng)
        return self.value


class Dirichlet(_MultiVariate):
    def __init__(self, par_ns, if_ignore_0s=False, id=None, name=None):
        """
        :param par_ns: (list) number of successes in each category
        :param if_ignore_0s: (bool) numpy requires all elements of par_ns to be >0. Setting this
            parameter to True allows par_ns to contain 0s.
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """

        _Parameter.__init__(self, id=id, name=name)
        self.parNs = par_ns
        self.ifIgnore0s = if_ignore_0s

    def sample(self, rng=None, time=None):

        self.value = Dirich(a=self.parNs, if_ignore_0s=self.ifIgnore0s).sample(rng=rng)
        return self.value


class AnOutcomeOfAMultiVariateDist(_SingleVariate):
    # a parameter that is defined on one outcome of a multi-variate parameter
    # such as multinomial or Dirichlet distributions
    def __init__(self, par_multivariate, outcome_index, id=None, name=None):
        """
        :param par_multivariate: (Parameter) a multivariate parameter such as multinomial or Dirichlet
        :param outcome_index: (int) index of the outcome of interest
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """

        assert isinstance(par_multivariate, _MultiVariate), \
            "par_multivariate should be multivariate parameter such as multinomial or Dirichlet. A {} is " \
            "provided instead.".format(type(par_multivariate))

        _Parameter.__init__(self, id=id, name=name)
        self.multivariate = par_multivariate
        self.i = outcome_index

    def sample(self, rng=None, time=None):
        self.value = self.multivariate.value[self.i]
        return self.value


class Inverse(_SingleVariate):
    # value = 1 / value of another parameter

    def __init__(self, par, id=None, name=None):
        """
        :param par: (Parameter) another parameter to use to calculate the inverse
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name, if_time_dep=par.ifTimeDep)
        self.par = par

    def sample(self, rng=None, time=None):
        self.value = 1/self.par.value
        return self.value


class OneMinus(_SingleVariate):
    # value = 1 - value of another parameter

    def __init__(self, par, id=None, name=None):
        """
        :param par: (Parameter) another parameter to use to calculate 1 - p
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name, if_time_dep=par.ifTimeDep)
        self.par = par

    def sample(self, rng=None, time=None):
        self.value = 1-self.par.value
        return self.value


class OneMinusTimes(_SingleVariate):
    # value = (1- par1) * par2

    def __init__(self, par1, par2, id=None, name=None):
        """
        :param par1: (Parameter) parameter 1
        :param par2: (Parameter) parameter 2
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name, if_time_dep=(par1.ifTimeDep or par2.ifTimeDep))
        self.par1 = par1
        self.par2 = par2

    def sample(self, rng=None, time=None):
        self.value = (1 - self.par1.value) * self.par2.value
        return self.value


class TenToPower(_SingleVariate):
    # 10^(value of another parameter)

    def __init__(self, par, id=None, name=None):
        """
        :param par: (Parameter) another parameter to use to calculate 10^p
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name, if_time_dep=par.ifTimeDep)
        self.par = par

    def sample(self, rng=None, time=None):
        self.value = pow(10, self.par.value)
        return self.value


class Logit(_SingleVariate):
    # p/(1-p)

    def __init__(self, par, id=None, name=None):
        """
        :param par: (Parameter) another parameter to use to calculate p(1 - p)
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name, if_time_dep=par.ifTimeDep)
        self.par = par

    def sample(self, rng=None, time=None):
        self.value = self.par.value/(1-self.par.value)
        return self.value


class ProbFromOdd(_SingleVariate):
    # return the probability such that the odd is k times the odd of the reference event
    # O = k*O0 or p = k*p0 / (1 + (k-1)*p0)

    def __init__(self, par_p_ref, par_multiplier, id=None, name=None):
        """
        :param par_p_ref: (Parameter) parameter the reference probability (p0)
        :param par_multiplier: (Parameter) parameter the multiplier for the odd of the reference (k)
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name, if_time_dep=(par_p_ref.ifTimeDep or par_multiplier.ifTimeDep))
        self.parPRef = par_p_ref
        self.parMultiplier = par_multiplier

    def sample(self, rng=None, time=None):
        self.value =  self.parMultiplier.value * self.parPRef.value/(1 + (self.parMultiplier.value-1)*self.parPRef.value)
        return self.value

class RateToOccur(_SingleVariate):
    """ determines rate of an event such that it occurs with certain probability during a certain period """
    def __init__(self, par_probability, delta_t, id=None, name=None):
        """
        :param par_probability: parameter for the probability that the event occurs during deltaT
        :param delta_t: time period over which the event should occur with the specified probability
        """

        _Parameter.__init__(self, id=id, name=name, if_time_dep=par_probability.ifTimeDep)
        self.parProb = par_probability
        self.deltaTInv = 1/delta_t

    def sample(self, rng=None, time=None):
        self.value = -log(1-self.parProb.value) * self.deltaTInv
        return self.value


class Division(_SingleVariate):
    # p1 / p2
    def __init__(self, par_numerator, par_denominator, id=None, name=None):
        """
        :param par_numerator: (Parameter) numerator parameter
        :param par_denominator: (Parameter) denominator parameter
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name, if_time_dep=(par_numerator.ifTimeDep or par_denominator.ifTimeDep))
        self.numerator = par_numerator
        self.denominator = par_denominator

    def sample(self, rng=None, time=None):
        self.value = self.numerator.value/self.denominator.value
        return self.value


class LinearCombination(_SingleVariate):
    # linear combination of multiple parameters

    def __init__(self, parameters, coefficients=None, id=None, name=None):
        """
        :param parameters: (list of Parameters)
        :param coefficients: (list of floats) list of coefficients
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """

        _Parameter.__init__(self, id=id, name=name,
                            if_time_dep=self.if_any_time_dep(parameters=parameters))
        self.parameters = parameters
        if coefficients is not None:
            self.coefficients = coefficients
        else:
            self.coefficients = [1] * len(parameters)

    def sample(self, rng=None, time=None):
        self.value = 0
        for i, p in enumerate(self.parameters):
            self.value += p.value * self.coefficients[i]

        return self.value


class OneMinusSum(_SingleVariate):
    # 1 - sum of of multiple parameters

    def __init__(self, parameters, id=None, name=None):
        """
        :param parameters: (list of Parameters)
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """

        # if this is a time-dependent parameter
        if_time_dep = False
        for p in parameters:
            if p.ifTimeDep:
                if_time_dep = True
                break

        _Parameter.__init__(self, id=id, name=name, if_time_dep=if_time_dep)
        self.params = parameters

    def sample(self, rng=None, time=None):

        self.value = 1
        for p in self.params:
            self.value -= p.value
        return self.value


class Product(_SingleVariate):
    # product of multiple parameters

    def __init__(self, parameters, id=None, name=None):
        """
        :param parameters: (list of Parameters)
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """

        _Parameter.__init__(self, id=id, name=name,
                            if_time_dep=self.if_any_time_dep(parameters=parameters))
        self.parameters = parameters

    def sample(self, rng=None, time=None):

        self.value = 1
        for p in self.parameters:
            if p.value is None:
                self.value = None
                break
            else:
                self.value *= p.value

        return self.value


class Surge(_SingleVariate):
    # f(t) = base.value * ( 1 + percentChange(t))
    # percentChange(t) = A * (1 - cos(2*pi*(t1-t)/(t1-t0)) / 2
    # A is maximum % change
    def __init__(self, par_base=1, par_max_percent_change=1, par_t0=0, par_t1=1, id=None, name=None):
        """
        :param par_base: (Parameter or float) value to use as base
        :param par_max_percent_change: (Parameter or float) maximum % change in base value
            (should be + for increase and - for decrease)
        :param par_t0: (Parameter or float) f(t) = 0 for t < t0
        :param par_t1: (Parameter or float) f(t) = 0 for t > t1
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """

        if not isinstance(par_base, _Parameter):
            par_base = Constant(value=par_base)
        if not isinstance(par_max_percent_change, _Parameter):
            par_max_percent_change = Constant(value=par_max_percent_change)
        if not isinstance(par_t0, _Parameter):
            par_t0 = Constant(value=par_t0)
        if not isinstance(par_t1, _Parameter):
            par_t1 = Constant(value=par_t1)

        _Parameter.__init__(self=self, id=id, name=name,
                            if_time_dep=True)

        self.base = par_base
        self.maxPercChange = par_max_percent_change
        self.t0 = par_t0
        self.t1 = par_t1
        self.twoPi = 2*pi

    def sample(self, rng=None, time=None):

        if time < self.t0.value or time > self.t1.value:
            self.value = self.base.value
        else:
            x = self.twoPi * (time-self.t0.value)/(self.t1.value - self.t0.value)
            percent_change = self.maxPercChange.value * (1 - cos(x)) / 2
            self.value = self.base.value * (1 + percent_change)
        return self.value


class SigmoidOnModelOutput(_SingleVariate):
    # f(h) = min + (max-min) * 1 / (1 + exp(-b * h))

    def __init__(self, par_b, par_min=None, par_max=None, id=None, name=None):
        """
        :param par_b: (Parameter) of b
        :param par_min: (Parameter) of min (if not provided, Constant(0) is used)
        :param par_max: (Parameter) of max (if not provided, Constant(1) is used)
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name, if_time_dep=True)

        self.parB = par_b
        self.parMin = par_min if par_min is not None else Constant(value=0)
        self.parMax = par_max if par_max is not None else Constant(value=1)
        self.simOut = None

    def assign_sim_output(self, sim_output):

        assert hasattr(sim_output, 'get_value'), 'sim_output must have the attribute .get_value()'

        self.simOut = sim_output

    def sample(self, rng=None, time=None):

        assert self.simOut is not None, 'Make sure that sim_out is assigned.'

        v = self.simOut.get_value()
        if v is None:
            self.value = 0
        else:
            logistic = 1 / (1 + exp(-self.parB.value * v))
            self.value = self.parMin.value + (self.parMax.value - self.parMin.value) * logistic

        return self.value


class TimeDependentSigmoid(_SingleVariate):
    # f(t) = min + (max-min) * 1 / (1 + exp(-b * (t - t_middle - t_min)) if t > t_min
    # returns min for t = -inf and max for t = inf if b >= 0

    def __init__(self, par_b, par_t_min=None, par_t_middle=None, par_min=None, par_max=None, id=None, name=None):
        """
        :param par_b: (Parameter) of b
        :param par_t_min: (Parameter) of t_min
        :param par_t_middle: (Parameter) of t_middle
        :param par_min: (Parameter) of min (if not provided, Constant(0) is used)
        :param par_max: (Parameter) of max (if not provided, Constant(1) is used)
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name, if_time_dep=True)

        self.parB = par_b
        self.parTMin = par_t_min if par_t_min is not None else Constant(value=0)
        self.parTMid = par_t_middle if par_t_middle is not None else Constant(value=0)
        self.parMin = par_min if par_min is not None else Constant(value=0)
        self.parMax = par_max if par_max is not None else Constant(value=1)

    def sample(self, rng=None, time=None):

        if time < self.parTMin.value:
            self.value = 0
        else:
            dt = time - self.parTMid.value - self.parTMin.value
            logistic = 1 / (1 + exp(-self.parB.value * dt))
            self.value = self.parMin.value + (self.parMax.value - self.parMin.value) * logistic

        return self.value


class TimeDependentCosine(_SingleVariate):
    # f(t) = a0 + a1 * Cos (2 * pi * (t - phase)/scale)

    def __init__(self, par_phase=None, par_scale=None, par_a0=None, par_a1=None, id=None, name=None):
        """
        :param par_phase: (Parameter) of phase
        :param par_scale: (Parameter) of scale
        :param par_a0: (Parameter) of base (if not provided, Constant(0) is used)
        :param par_a1: (Parameter) of a (if not provided, Constant(1) is used)
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name, if_time_dep=True)

        self.parPhase = par_phase
        self.parScale = par_scale
        self.parA0 = par_a0 if par_a0 is not None else Constant(value=0)
        self.parA1 = par_a1 if par_a1 is not None else Constant(value=1)

    def sample(self, rng=None, time=None):

        arg = (time - self.parPhase.value)/self.parScale.value
        cosine = cos(2*pi*arg)
        self.value = self.parA0.value + self.parA1.value * cosine

        return self.value


class TimeDependentStepWise(_SingleVariate):
    # f(t) = 0  for       t < t0
    #      = v0 for t0 <= t < t1
    #      = v1 for t1 <= t < t2
    #      = v2 for t2 <= t

    def __init__(self, ts, vs, id=None, name=None):
        """
        :param ts: (list) of time break points
        :param vs: (list) of values
        :param id: (int) id of a parameter
        :param name: (string) name of a parameter
        """
        _Parameter.__init__(self, id=id, name=name, if_time_dep=True)

        assert len(ts) == len(vs), 'There should be an equal number of time breakpoints (ts) and function values (vs).'

        self.ts = ts
        self.vs = vs

    def sample(self, rng=None, time=None):

        self.value = 0
        if time < self.ts[0]:
            self.value = 0
        else:
            i = 0
            while i+1 < len(self.ts):
                if time < self.ts[i+1]:
                    break
                else:
                    i += 1

            self.value = self.vs[i]

        return self.value


class MatrixOfParams(_SingleVariate):
    def __init__(self, matrix_of_params_or_values, id=None, name=None):
        """
        :param matrix_of_params_or_values: (list of list) of numbers or Parameters
        :param id:
        :param name:
        """

        self.matrixOfParams = []
        for row in matrix_of_params_or_values:
            params = []
            for v in row:
                if isinstance(v, _Parameter):
                    params.append(v)
                else:
                    params.append(Constant(value=v))
            self.matrixOfParams.append(params)

        # find if time-dependant
        if_time_dep = False
        for row in self.matrixOfParams:
            for v in row:
                if v.ifTimeDep:
                    if_time_dep = True
                    break
            if if_time_dep:
                break

        _Parameter.__init__(self, id=id, name=name, if_time_dep=if_time_dep)

        if not self.ifTimeDep:
            self.sample()

    def sample(self, rng=None, time=None):
        """
        :return: (np.array)
        """

        self.value = []
        for params in self.matrixOfParams:
            values = []
            for par in params:
                values.append(par.value)
            self.value.append(values)

        self.value = np.array(self.value)
        return self.value


class ValuesOfParams(_MultiVariate):
    """ returns the values of a list of parameters """

    def __init__(self, parameters, id=None, name=None):
        """
        :param parameters: (list) of Parameters
        :param id:
        :param name:
        """

        self.parameters = parameters
        # find if time-dependant
        if_time_dep = False
        for p in self.parameters:
            if p.ifTimeDep:
                if_time_dep = True
                break
        _Parameter.__init__(self, id=id, name=name, if_time_dep=if_time_dep)

    def sample(self, rng=None, time=None):

        self.value = []
        for param in self.parameters:
            self.value.append(param.value)
        return self.value

